/*
VODKA-INDUCED ENTERTAINMENT ADVANCED SOUND MANAGER v0.9.1
COMPATIBILITY LAYER FOR WOLF4SDL v1.6
BY GERARD 'ALUMIUN' WATSON

MODIFIED BY WOLF 'WSJ' SKEVOS-JONES

This file replaces all external ID_SD functions and audio-based ID_CA
functions in Wolf4SDL v1.6 for use with the Advanced Sound Manager.
*/

#include "wl_def.h"
#include "asmcref.h"

byte   SoundMode,MusicMode,soundvol,musicvol,reversestereo,curmusic;

sample      ASM_Audiosegs[NUMSOUNDS];
chanstruct  channelinfo[ASM_MAXCHANNELS];


/*
=============================================================================

                              SOUND EFFECTS

=============================================================================
*/

/*
==========================
=
= SetSoundLoc - Given the location of an object (in terms of global
=       coordinates, held in globalsoundx and globalsoundy), munges the values
=       for an approximate distance from the left and right ear, and puts
=       those values into leftchannel and rightchannel.
=
= JAB
=
==========================
*/

#define ATABLEMAX 15
byte righttable[ATABLEMAX][ATABLEMAX * 2] = {
{ 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 7, 7, 6, 0, 0, 0, 0, 0, 1, 3, 5, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 7, 6, 4, 0, 0, 0, 0, 0, 2, 4, 6, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 6, 6, 4, 1, 0, 0, 0, 1, 2, 4, 6, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 7, 6, 5, 4, 2, 1, 0, 1, 2, 3, 5, 7, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 6, 5, 4, 3, 2, 2, 3, 3, 5, 6, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 6, 6, 5, 4, 4, 4, 4, 5, 6, 7, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 6, 6, 5, 5, 5, 6, 6, 7, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 7, 7, 7, 6, 6, 7, 7, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8}
};
byte lefttable[ATABLEMAX][ATABLEMAX * 2] = {
{ 8, 8, 8, 8, 8, 8, 8, 8, 5, 3, 1, 0, 0, 0, 0, 0, 6, 7, 7, 7, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 6, 4, 2, 0, 0, 0, 0, 0, 4, 6, 7, 7, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 6, 4, 2, 1, 0, 0, 0, 1, 4, 6, 6, 7, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 7, 5, 3, 2, 1, 0, 1, 2, 4, 5, 6, 7, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 6, 5, 3, 3, 2, 2, 3, 4, 5, 6, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 7, 6, 5, 4, 4, 4, 4, 5, 6, 6, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 7, 6, 6, 5, 5, 5, 6, 6, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 7, 7, 6, 6, 7, 7, 7, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8},
{ 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 8}
};

void SD_SetSoundLoc(int channel, fixed gx, fixed gy)
{
    fixed   xt,yt;
    int     x,y,leftvol,rightvol,distance;

    if (!Mix_Playing(channel))
        return;

    channelinfo[channel].x = gx;
    channelinfo[channel].y = gy;

    if (gx < 1 || gy < 1)    // WSJ: non-locational sound?
    {
        Mix_SetPosition(channel, 0, 0);
        if (reversestereo)
            Mix_SetReverseStereo(channel, 1);
        else
            Mix_SetReverseStereo(channel, 0);
        return;
    }

//
// WSJ: hack to make up for the crappy distance attenuation
//
    distance = (GlobalDist(player->x,player->y,gx,gy) >> (TILESHIFT - 1));

    if (distance > 127)
    {
        SD_StopChannel(channel);    // WSJ: too far away
        return;
    }
    if (distance < 7)
        distance = 0;

//
// translate point to view centered coordinates
//
    gx -= viewx;
    gy -= viewy;

//
// calculate newx
//
    xt = FixedMul(gx,viewcos);
    yt = FixedMul(gy,viewsin);
    x = (xt - yt) >> TILESHIFT;

//
// calculate newy
//
    xt = FixedMul(gx,viewsin);
    yt = FixedMul(gy,viewcos);
    y = (yt + xt) >> TILESHIFT;

    if (y >= ATABLEMAX)
        y = ATABLEMAX - 1;
    else if (y <= -ATABLEMAX)
        y = -ATABLEMAX;
    if (x < 0)
        x = -x;
    if (x >= ATABLEMAX)
        x = ATABLEMAX - 1;

    leftvol = ((15 - lefttable[x][y + ATABLEMAX]) << 4) + 15 - distance;
    rightvol = ((15 - righttable[x][y + ATABLEMAX]) << 4) + 15 - distance;

    Mix_SetPanning(channel,leftvol,rightvol);
}

void SD_UpdateSoundLoc(void)
{
    for (int i = 0; i < ASM_MAXCHANNELS; i++)
    {
        SD_SetSoundLoc(i,channelinfo[i].x,channelinfo[i].y);
    }
}

/*
**      JAB End
*/

void SD_ChannelDone(int channel)
{
    memset(&channelinfo[channel],0,sizeof(chanstruct));
}

void SD_Startup(void)
{
    if (nosound)
        return;

    if (!ASM_Open(param_samplerate, param_8bitsound, param_audiobuffer, 128, 128, false))
    {
        printf ("SD_Startup: Unable to open audio device. Trying defaults.\n"); 
        if (!ASM_Open(22050, true, 2048, 128, 128, false))
        {    
            printf ("SD_Startup: Unable to open audio device with defaults.\n"
                    "            Try closing any other applications that might\n"
                    "            be using the audio driver and try again, or\n"
                    "            start the program with the '--nosound' switch.\n");
            exit(1);
        }
    }
    
    Mix_ChannelFinished(SD_ChannelDone);
    memset (ASM_Audiosegs, 0, sizeof(ASM_Audiosegs));
}

void SD_Shutdown(void)
{
    for (int i = 0; i < NUMSOUNDS; i++)
        if (ASM_Audiosegs[i].chunk != NULL)
            UNCACHEAUDIOCHUNK(i);

    ASM_Close();
}

// WSJ: changed to void because it doesn't really need to return anything
void SD_CacheAudioChunk(int chunk)
{
    char *file;

    if (chunk >= NUMSOUNDS)
    {
#ifdef VERBOSE
        printf("SD_CacheAudioChunk: Invalid sample number %d!", chunk);
#endif
        return;
    }

    file = CA_GetFilePath(ASM_Soundnames[chunk],SOUNDDIR);
    if (file != NULL)
    {
        ASM_Audiosegs[chunk] = ASM_Cache(file, ASM_Soundnames[chunk]);
        free (file);
    }
}

void SD_LoadAllSounds(void)
{
    ASM_AbortIfClosed;
    SD_AbortIfSndOff;

    for (int i = 0; i < NUMSOUNDS; i++)
    {
        if (ASM_Audiosegs[i].chunk != NULL)
            continue;
        SD_CacheAudioChunk(i);
    }
}

void SD_StopChannel(int channel)
{
    ASM_AbortIfClosed;

    Mix_HaltChannel(channel);
}

void SD_PauseSound(void)
{
    ASM_AbortIfClosed;

    Mix_Pause(-1);
}

void SD_ResumeSound(void)
{
    ASM_AbortIfClosed;
    SD_AbortIfSndOff;

    Mix_Resume(-1);
}

int SD_ChannelPlaying(int channel)
{
    ASM_AbortIfClosed 0;
    SD_AbortIfSndOff 0;

    return Mix_Playing(channel);
}

bool SD_ChannelLooping(int channel)
{
    ASM_AbortIfClosed false;
    SD_AbortIfSndOff false;

    return (channelinfo[channel].flags&SNDFL_LOOPED && Mix_Playing(channel));
}

void SD_SetSoundMode(byte mode)
{
    if (mode == SoundMode)
        return;

    SD_StopSound();

    if (!mode)
        SoundMode = 0;
    else
        SoundMode = 1;
}

// WSJ: returns the channel which the specified sound is on, if it's playing
int SD_SoundPlaying(word sound, int channel)
{
    ASM_AbortIfClosed -1;
    SD_AbortIfSndOff -1;

    if (channel > -1 && channel < ASM_MAXCHANNELS)
    {
        if (Mix_Playing(channel) && channelinfo[channel].cursound==sound)
            return channel;
        return -1;
    }

    for (int i = 0; i < ASM_MAXCHANNELS; i++)
    {
        if (Mix_Playing(i) && channelinfo[i].cursound==sound)
            return i;
    }

    return -1;
}

// WSJ: returns true if the sound can be played on the specified channel
bool SD_CheckChannel(int channel, word sound, fixed x, fixed y, byte flags)
{
    fixed dist1,dist2;

    if (Mix_Playing(channel))
    {
        if (flags & SNDFL_CHECKDONE && SD_SoundPlaying(sound,channel) != -1)
            return false;    // WSJ: still playing, so don't repeat yet

        if (flags & SNDFL_CHECKDIST)
        {
            if (channelinfo[channel].x == 0 || channelinfo[channel].y == 0)
                return false;    // WSJ: sound is further away than the current one

            dist1 = GlobalDist(x,y,player->x,player->y);
            if ((dist1 >> (TILESHIFT - 2)) > 127)
                return false;

            dist2 = GlobalDist(channelinfo[channel].x,channelinfo[channel].y,player->x,player->y);
            if (dist1 > dist2)
                return false;    // WSJ: sound is further away than the current one
        }
    }

    return true;
}

int SD_PlaySound(word sound, int channel, fixed x, fixed y, byte flags)
{
    ASM_AbortIfClosed -1;
    SD_AbortIfSndOff -1;

    if (sound >= NUMSOUNDS)
    {
#ifdef VERBOSE
        printf ("SD_PlaySound: Invalid sound number %d!\n", sound);
#endif
        return -1;
    }

    if (ASM_Audiosegs[sound].chunk == NULL)
    {
#ifdef VERBOSE
        printf ("SD_PlaySound: Sound %d not cached!\n", sound);
#endif
        return -1;
    }

    if (x < 1 || y < 1)
        flags &= ~SNDFL_CHECKDIST;    // WSJ: sound will be heard everywhere

    if (!SD_CheckChannel(channel,sound,x,y,flags))
        return -1;

    if (flags & SNDFL_LOOPED)
    {
        if (flags & SNDFL_CHECKDIST && SD_SoundPlaying(sound,channel) != -1)
        {
            SD_SetSoundLoc(channel,x,y);
            return -1;    // WSJ: update looped sound's location, but don't restart it
        }
        channel = ASM_PlaySound(ASM_Audiosegs[sound],channel,true);
    }
    else
        channel = ASM_PlaySound(ASM_Audiosegs[sound],channel,false);

    if (channel > -1)
    {
        memset(&channelinfo[channel],0,sizeof(chanstruct));
        channelinfo[channel].cursound = sound;
        channelinfo[channel].flags = flags;
        SD_SetSoundLoc(channel,x,y);
    }
    return channel;
}


/*
=============================================================================

                              MUSIC

=============================================================================
*/

int SD_MusicPlaying(int song)
{
    ASM_AbortIfClosed 0;
    SD_AbortIfMusOff 0;

    if (song < 1 || curmusic != song)
        return 0;

    return Mix_PlayingMusic();
}

void SD_PauseMusic(void)
{
    ASM_AbortIfClosed;

    Mix_PauseMusic();
}

void SD_ResumeMusic(void)
{
    ASM_AbortIfClosed;
    SD_AbortIfMusOff;

    Mix_ResumeMusic();
}

void SD_Pause(void)
{
    SD_PauseSound();
    SD_PauseMusic();
}

void SD_Resume(void)
{
    SD_ResumeSound();
    SD_ResumeMusic();
}

void SD_StopMusic(void)
{
    ASM_AbortIfClosed;

    curmusic = 0;
    Mix_HaltMusic();
}

// WSJ: play music track if it exists
void SD_StartMusic(int song)
{
    ASM_AbortIfClosed;
    SD_AbortIfMusOff;

    char *file;
    char filename[12];

    if (song > 0 && song < 100)
    {
        strcpy (filename,"track00.ogg");
        filename[5] = '0' + (song/10);
        filename[6] = '0' + (song%10);
        file = CA_GetFilePath(filename,MUSICDIR);
        if (file != NULL)
        {
            bool result = ASM_PlayMusic(file);
            free (file);

            if (result)
            {
                curmusic = song;
                return;
            }
        }
    }

    SD_StopMusic();
}

void SD_SetMusicMode(byte mode)
{
    if (mode == MusicMode)
        return;

    if (!mode)
    {
        MusicMode = 0;
        SD_PauseMusic();
    }
    else
    {
        MusicMode = 1;
        if (!ingame)
            SD_ResumeMusic();
    }
}
