/*
VODKA-INDUCED ENTERTAINMENT ADVANCED SOUND MANAGER v0.9.1
BY GERARD 'ALUMIUN' WATSON

MODIFIED BY WOLF 'WSJ' SKEVOS-JONES
*/

// Initialization stuff

#include "vieasm.h"                    // Duh.
#include <SDL.h>

Uint8 sndvol, musvol;                       // Volumes for sound
static Mix_Music *music = 0;                // Music references
bool deviceopen = false;                    // Is device open?
bool reversemode = false;                   // Reverse stereomode

// ASM_IsOpen
// Returns true if the device is open, false otherwise

bool ASM_IsOpen(void)
{
    return (deviceopen) ? true : false;
}

// ASM_Open
// Opens audio device at given specs, clears used

bool ASM_Open(int frequency, bool use8bit, int buffersize, Uint8 sndvolume, Uint8 musvolume, bool reverse)
{
    int value;

    if (ASM_IsOpen())       // Device is already open!
        return false;

    if (use8bit)
        value = Mix_OpenAudio(frequency, AUDIO_S8, 2, buffersize);
    else
        value = Mix_OpenAudio(frequency, AUDIO_S16, 2, buffersize);

    if(value == -1)
    {
#ifdef VERBOSE
        printf("ASM_Open: %s\n", Mix_GetError());
#endif
        return false;
    }

    Mix_AllocateChannels(ASM_MAXCHANNELS);

    if (reverse)
        reversemode = true;

    deviceopen = true;

    ASM_ChangeVolume(sndvolume, musvolume);
    Mix_ChannelFinished(NULL);
    return true;
}

// ASM_Close
// Closes the audio device

void ASM_Close(void)
{
    if (music)
    {
        Mix_HookMusicFinished(NULL);
        Mix_HaltMusic();
        Mix_FreeMusic(music);
    }

#ifdef VERBOSE
    printf("ASM_Close: Audio device closed.\n");
#endif
    Mix_CloseAudio();
    deviceopen = false;
}

// ASM_PlayMusic
// Plays a music sample from a file

bool ASM_PlayMusic(char *musfile)
{
    ASM_AbortIfClosed false;

    if (Mix_PlayingMusic())
        Mix_HaltMusic();

    Mix_HookMusicFinished(NULL);

    if (music)
        Mix_FreeMusic(music);

    music = Mix_LoadMUS(musfile);
    if (!music)
        return false;

    Mix_VolumeMusic(musvol);

    if (Mix_FadeInMusic(music, -1, 0) == -1)
    {
#ifdef VERBOSE
        printf("ASM_PlayMusic: %s\n", Mix_GetError());
#endif
        return false;
    }

    return true;
}

// ASM_ChangeVolume
// Changes the volume of:
//      Sound - iterates through each channel and sets volume accordingly
//      Music - just calls Mix_VolumeMusic

void ASM_ChangeVolume(Uint8 sndvolume, Uint8 musvolume)
{
    ASM_AbortIfClosed;

    sndvol = (sndvolume > MIX_MAX_VOLUME) ? MIX_MAX_VOLUME : sndvolume;
    musvol = (musvolume > MIX_MAX_VOLUME) ? MIX_MAX_VOLUME : musvolume;
    Mix_Volume(-1, sndvol);
    Mix_VolumeMusic(musvol);
}

// ASM_Cache
// Loads a sound into memory and returns it as a sample

sample ASM_Cache(char *sndfile, const char *name)
{
    sample sound;
    sound.chunk = NULL;
    sound.name = NULL;

    ASM_AbortIfClosed sound;

    Mix_Chunk *chunk = Mix_LoadWAV(sndfile);

    if (chunk == NULL)
    {
#ifdef VERBOSE
        printf("ASM_Cache: %s on sound %s\n", Mix_GetError(), sndfile);
#endif
        return sound;
    }

    sound.chunk = chunk;
    sound.name = (char *)malloc((size_t)strlen(name)+1);
    strcpy (sound.name, name);

    return sound;
}

// ASM_Uncache
// Removes a sound from memory

void ASM_Uncache(sample &sound)
{
    ASM_AbortIfClosed;
    
    if (sound.chunk == NULL)
        return;

    Mix_FreeChunk(sound.chunk);
    sound.chunk = NULL;
    free(sound.name);
    sound.name = NULL;
}

// ASM_PlaySound
// Play a sound in memory loaded with ASM_Cache or Mix_LoadWAV
// Returns -1 on errors or channel number

int ASM_PlaySound(sample &sound, int channel, bool looped)
{
    ASM_AbortIfClosed -1;

    if (sound.chunk == NULL)
    {
#ifdef VERBOSE
        printf("ASM_PlaySound: \"%s\" not cached!\n", sound.name);
#endif
        return -1;
    }

    if (channel<0 || channel>=ASM_MAXCHANNELS)
        return -1;

    Mix_Volume (channel, sndvol);
    
    if (Mix_PlayChannel(channel, sound.chunk, (looped) ? -1 : 0) == -1)
    {
#ifdef VERBOSE
        printf("ASM_PlaySound: %s\n", Mix_GetError());
#endif
        return -1;
    }

    Mix_SetPosition(channel, 0, 0);
    if (reversemode)
        Mix_SetReverseStereo(channel, 1);
    else
        Mix_SetReverseStereo(channel, 0);

    return channel;
}
        
// ASM_ReverseStereo
// Sets reverse stereo mode after initialization

void ASM_ReverseStereo(bool reverse)
{
    ASM_AbortIfClosed;

    Mix_SetReverseStereo(MIX_CHANNEL_POST, (reverse) ? 1 : 0);
}
