#ifndef WL_DEF_H
#define WL_DEF_H

// Defines which version shall be built and configures supported extra features
#include "version.h"

#include <assert.h>
#include <fcntl.h>
#include <math.h>
#include <time.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#if defined(_arch_dreamcast)
#	include <kos.h>
#elif !defined(_WIN32)
#	include <stdint.h>
#	include <string.h>
#	include <stdarg.h>
#endif
#include <SDL.h>

#if !defined O_BINARY
#	define O_BINARY 0
#endif

#pragma pack(1)

#include "foreign.h"

#ifndef SPEAR
    #include "audiowl6.h"
    #ifdef UPLOAD
        #include "gfxv_apo.h"
    #else
        #ifdef GOODTIMES
            #include "gfxv_wl6.h"
        #else
            #include "gfxv_apo.h"
        #endif
    #endif
#else
    #include "audiosod.h"
    #include "gfxv_sod.h"
    #include "f_spear.h"
#endif

#ifdef _arch_dreamcast
typedef uint8 uint8_t;
typedef uint16 uint16_t;
typedef uint32 uint32_t;
typedef int8 int8_t;
typedef int16 int16_t;
typedef int32 int32_t;
typedef int64 int64_t;
typedef ptr_t uintptr_t;
#endif

typedef uint8_t byte;
typedef uint16_t word;
typedef int32_t fixed;
typedef uint32_t longword;
typedef int8_t boolean;
typedef void * memptr;

typedef struct
{
    int x,y;
} Point;
typedef struct
{
    Point ul,lr;
} Rect;

void Quit(const char *errorStr, ...);

#include "id_pm.h"
#include "asmcomp.h"
#include "id_in.h"
#include "id_vl.h"
#include "id_vh.h"
#include "id_us.h"
#include "id_ca.h"

#include "wl_menu.h"

#ifdef SPEAR
#include "f_spear.h"
#endif

#define MAPSPOT(x,y,plane) (mapsegs[plane][((y)<<mapshift)+(x)])

#define SIGN(x)         ((x)>0?1:-1)
#define ABS(x)          ((int)(x)>0?(x):-(x))
#define LABS(x)         ((int32_t)(x)>0?(x):-(x))

#define abs(x) ABS(x)

#define TILEDIST(tx1,ty1,tx2,ty2) ((int)abs(tx1-tx2)>abs(ty1-ty2) ? abs(tx1-tx2) : abs(ty1-ty2))
#define OBJTILEDIST(ob1,ob2) TILEDIST(ob1->tilex,ob1->tiley,ob2->tilex,ob2->tiley)
#define TILECENTER(tn)  (((int32_t)tn<<TILESHIFT)+TILEGLOBAL/2)

/*
=============================================================================

                            GLOBAL CONSTANTS

=============================================================================
*/

#define MAXTICS 10
#define DEMOTICS        4

#define MAXACTORS       1000        // max number of nazis, etc / map
#define MAXSTATS        1000        // max number of lamps, bonus, etc
#define MAXDOORS        128         // max number of sliding doors
#define MAXWALLTILES    77          // max number of wall tiles
#define FIRSTDOORSIDE   (MAXWALLTILES<<1)
#define DOORWALL        (FIRSTDOORSIDE+4)

//
// tile constants
//

#define ICONARROWS      90
#define PUSHABLETILE    98
#define EXITTILE        99     // at end of castle
#define ENDLEVELTILE    100
#define ENEMYBLOCK      101    // WSJ: blocks enemies who shoot, but not dogs
#define ENEMYBLOCK2     102    // WSJ: blocks all enemies
#define KEY1ENEMYBLOCK  103    // WSJ: blocks all enemies until player has key
#define KEY2ENEMYBLOCK  104
#define KEY3ENEMYBLOCK  105
#define KEY4ENEMYBLOCK  106
#define ACIDTILE        225

#define AREATILE        107    // first of NUMAREAS floor tiles
#define NUMAREAS        41
#define LASTAREA        (AREATILE+NUMAREAS-1)
#define ELEVATORTILE    21
#define BLOCKTILE       64
#define LANDSCAPETILE   74
#define AMBUSHTILE      106
#define SECRETEXIT1TILE 144
#define SECRETEXIT2TILE 145
#define SECRETEXIT3TILE 146
#define DORMANT1TILE    147    // WSJ: dormant enemy (LASTAREA, stays on map)
#define DORMANT2TILE    148    // WSJ: dormant enemy (deleted like AMBUSHTILE)

#define NUMBERCHARS     9


//----------------

#define EXTRAPOINTS     40000

#define PLAYERSPEED     3000
#define RUNSPEED        6000

#define SCREENSEG       0xa000

#define SCREENBWIDE     80

#define HEIGHTRATIO     0.50            // also defined in id_mm.c

#define BORDERCOLOR     3
#define FLASHCOLOR      5
#define FLASHTICS       4

#define LRpack          17      // # of levels to store in endgame
#define NUMBONUSMAPS    4       // WSJ: # of levels NOT stored in endgame

#define PLAYERSIZE      MINDIST         // player radius
#define MINACTORDIST    0x10000l        // minimum dist from player center
                                        // to any actor center
#define TOTENDELAYTICS  500

#define NUMLATCHPICS    100

#undef M_PI
#define PI              3.141592657
#define M_PI PI

#define GLOBAL1         (1l<<16)
#define TILEGLOBAL      GLOBAL1
#define PIXGLOBAL       (GLOBAL1/64)
#define TILESHIFT       16l
#define UNSIGNEDSHIFT   8

#define ANGLES          360             // must be divisable by 4
#define ANGLEQUAD       (ANGLES/4)
#define FINEANGLES      3600
#define ANG90           (FINEANGLES/4)
#define ANG180          (ANG90*2)
#define ANG270          (ANG90*3)
#define ANG360          (ANG90*4)
#define VANG90          (ANGLES/4)
#define VANG180         (VANG90*2)
#define VANG270         (VANG90*3)
#define VANG360         (VANG90*4)

#define MINDIST         (0x5800l)

#define mapshift        6
#define MAPSIZE         (1<<mapshift)
#define maparea         MAPSIZE*MAPSIZE

#define mapheight       MAPSIZE
#define mapwidth        MAPSIZE

#ifdef USE_HIRES

#define TEXTURESHIFT    7
#define TEXTURESIZE     (1<<TEXTURESHIFT)
#define TEXTUREFROMFIXEDSHIFT 2
#define TEXTUREMASK     (TEXTURESIZE*(TEXTURESIZE-1))

#define SPRITESCALEFACTOR 1

#else

#define TEXTURESHIFT    6
#define TEXTURESIZE     (1<<TEXTURESHIFT)
#define TEXTUREFROMFIXEDSHIFT 4
#define TEXTUREMASK     (TEXTURESIZE*(TEXTURESIZE-1))

#define SPRITESCALEFACTOR 2

#endif

#define NORTH   0
#define EAST    1
#define SOUTH   2
#define WEST    3


#define STATUSLINES     40

#define SCREENSIZE      (SCREENBWIDE*208)
#define PAGE1START      0
#define PAGE2START      (SCREENSIZE)
#define PAGE3START      (SCREENSIZE*2u)
#define FREESTART       (SCREENSIZE*3u)


#define PIXRADIUS       512

#define NUMFLAKGUNPARTS 3


// object flag values

typedef enum
{
    FL_SHOOTABLE        = 0x00000001,
    FL_BONUS            = 0x00000002,
    FL_NEVERMARK        = 0x00000004,
    FL_VISABLE          = 0x00000008,
    FL_ATTACKMODE       = 0x00000010,
    FL_FIRSTATTACK      = 0x00000020,
    FL_AMBUSH           = 0x00000040,
    FL_NONMARK          = 0x00000080,
    FL_USED             = 0x00000080,    // WSJ: item dropped by enemy
    FL_FULLBRIGHT       = 0x00000100,
#ifdef USE_DIR3DSPR
    // you can choose one of the following values in wl_act1.cpp
    // to make a static sprite a directional 3d sprite
    // (see example at the end of the statinfo array)
    FL_DIR_HORIZ_MID    = 0x00000200,
    FL_DIR_HORIZ_FW     = 0x00000400,
    FL_DIR_HORIZ_BW     = 0x00000600,
    FL_DIR_VERT_MID     = 0x00000a00,
    FL_DIR_VERT_FW      = 0x00000c00,
    FL_DIR_VERT_BW      = 0x00000e00,

    // these values are just used to improve readability of code
    FL_DIR_NONE         = 0x00000000,
    FL_DIR_POS_MID      = 0x00000200,
    FL_DIR_POS_FW       = 0x00000400,
    FL_DIR_POS_BW       = 0x00000600,
    FL_DIR_POS_MASK     = 0x00000600,
    FL_DIR_VERT_FLAG    = 0x00000800,
    FL_DIR_MASK         = 0x00000e00,
#endif
} objflag_t;


//
// sprite constants
//

enum
{
    SPR_LSCOPESIDE,SPR_RSCOPESIDE,

//
// static sprites
//
    SPR_STAT_0,SPR_STAT_1,SPR_STAT_2,SPR_STAT_3,
    SPR_STAT_4,SPR_STAT_5,SPR_STAT_6,SPR_STAT_7,

    SPR_STAT_8,SPR_STAT_9,SPR_STAT_10,SPR_STAT_11,
    SPR_STAT_12,SPR_STAT_13,SPR_STAT_14,SPR_STAT_15,

    SPR_STAT_16,SPR_STAT_17,SPR_STAT_18,SPR_STAT_19,
    SPR_STAT_20,SPR_STAT_21,SPR_STAT_22,SPR_STAT_23,

    SPR_STAT_24,SPR_STAT_25,SPR_STAT_26,SPR_STAT_27,
    SPR_STAT_28,SPR_STAT_29,SPR_STAT_30,SPR_STAT_31,

    SPR_STAT_32,SPR_STAT_33,SPR_STAT_34,SPR_STAT_35,
    SPR_STAT_36,SPR_STAT_37,SPR_STAT_38,SPR_STAT_39,

    SPR_STAT_40,SPR_STAT_41,SPR_STAT_42,SPR_STAT_43,
    SPR_STAT_44,SPR_STAT_45,SPR_STAT_45b,SPR_STAT_47,

//
// guard
//
    SPR_GRD_S_1,SPR_GRD_S_2,SPR_GRD_S_3,SPR_GRD_S_4,
    SPR_GRD_S_5,SPR_GRD_S_6,SPR_GRD_S_7,SPR_GRD_S_8,

    SPR_GRD_W1_1,SPR_GRD_W1_2,SPR_GRD_W1_3,SPR_GRD_W1_4,
    SPR_GRD_W1_5,SPR_GRD_W1_6,SPR_GRD_W1_7,SPR_GRD_W1_8,

    SPR_GRD_W2_1,SPR_GRD_W2_2,SPR_GRD_W2_3,SPR_GRD_W2_4,
    SPR_GRD_W2_5,SPR_GRD_W2_6,SPR_GRD_W2_7,SPR_GRD_W2_8,

    SPR_GRD_W3_1,SPR_GRD_W3_2,SPR_GRD_W3_3,SPR_GRD_W3_4,
    SPR_GRD_W3_5,SPR_GRD_W3_6,SPR_GRD_W3_7,SPR_GRD_W3_8,

    SPR_GRD_W4_1,SPR_GRD_W4_2,SPR_GRD_W4_3,SPR_GRD_W4_4,
    SPR_GRD_W4_5,SPR_GRD_W4_6,SPR_GRD_W4_7,SPR_GRD_W4_8,

    SPR_GRD_PAIN_1,SPR_GRD_DIE_1,SPR_GRD_DIE_2,SPR_GRD_DIE_3,
    SPR_GRD_PAIN_2,SPR_GRD_DEAD,

    SPR_GRD_SHOOT1,SPR_GRD_SHOOT2,SPR_GRD_SHOOT3,

//
// dogs
//
    SPR_DOG_W1_1,SPR_DOG_W1_2,SPR_DOG_W1_3,SPR_DOG_W1_4,
    SPR_DOG_W1_5,SPR_DOG_W1_6,SPR_DOG_W1_7,SPR_DOG_W1_8,

    SPR_DOG_W2_1,SPR_DOG_W2_2,SPR_DOG_W2_3,SPR_DOG_W2_4,
    SPR_DOG_W2_5,SPR_DOG_W2_6,SPR_DOG_W2_7,SPR_DOG_W2_8,

    SPR_DOG_W3_1,SPR_DOG_W3_2,SPR_DOG_W3_3,SPR_DOG_W3_4,
    SPR_DOG_W3_5,SPR_DOG_W3_6,SPR_DOG_W3_7,SPR_DOG_W3_8,

    SPR_DOG_W4_1,SPR_DOG_W4_2,SPR_DOG_W4_3,SPR_DOG_W4_4,
    SPR_DOG_W4_5,SPR_DOG_W4_6,SPR_DOG_W4_7,SPR_DOG_W4_8,

    SPR_DOG_DIE_1,SPR_DOG_DIE_2,SPR_DOG_DIE_3,SPR_DOG_DEAD,
    SPR_DOG_JUMP1,SPR_DOG_JUMP2,SPR_DOG_JUMP3,

//
// ss
//
    SPR_SS_S_1,SPR_SS_S_2,SPR_SS_S_3,SPR_SS_S_4,
    SPR_SS_S_5,SPR_SS_S_6,SPR_SS_S_7,SPR_SS_S_8,

    SPR_SS_W1_1,SPR_SS_W1_2,SPR_SS_W1_3,SPR_SS_W1_4,
    SPR_SS_W1_5,SPR_SS_W1_6,SPR_SS_W1_7,SPR_SS_W1_8,

    SPR_SS_W2_1,SPR_SS_W2_2,SPR_SS_W2_3,SPR_SS_W2_4,
    SPR_SS_W2_5,SPR_SS_W2_6,SPR_SS_W2_7,SPR_SS_W2_8,

    SPR_SS_W3_1,SPR_SS_W3_2,SPR_SS_W3_3,SPR_SS_W3_4,
    SPR_SS_W3_5,SPR_SS_W3_6,SPR_SS_W3_7,SPR_SS_W3_8,

    SPR_SS_W4_1,SPR_SS_W4_2,SPR_SS_W4_3,SPR_SS_W4_4,
    SPR_SS_W4_5,SPR_SS_W4_6,SPR_SS_W4_7,SPR_SS_W4_8,

    SPR_SS_PAIN_1,SPR_SS_DIE_1,SPR_SS_DIE_2,SPR_SS_DIE_3,
    SPR_SS_PAIN_2,SPR_SS_DEAD,

    SPR_SS_SHOOT1,SPR_SS_SHOOT2,SPR_SS_SHOOT3,

//
// mutant
//
    SPR_MUT_S_1,SPR_MUT_S_2,SPR_MUT_S_3,SPR_MUT_S_4,
    SPR_MUT_S_5,SPR_MUT_S_6,SPR_MUT_S_7,SPR_MUT_S_8,

    SPR_MUT_W1_1,SPR_MUT_W1_2,SPR_MUT_W1_3,SPR_MUT_W1_4,
    SPR_MUT_W1_5,SPR_MUT_W1_6,SPR_MUT_W1_7,SPR_MUT_W1_8,

    SPR_MUT_W2_1,SPR_MUT_W2_2,SPR_MUT_W2_3,SPR_MUT_W2_4,
    SPR_MUT_W2_5,SPR_MUT_W2_6,SPR_MUT_W2_7,SPR_MUT_W2_8,

    SPR_MUT_W3_1,SPR_MUT_W3_2,SPR_MUT_W3_3,SPR_MUT_W3_4,
    SPR_MUT_W3_5,SPR_MUT_W3_6,SPR_MUT_W3_7,SPR_MUT_W3_8,

    SPR_MUT_W4_1,SPR_MUT_W4_2,SPR_MUT_W4_3,SPR_MUT_W4_4,
    SPR_MUT_W4_5,SPR_MUT_W4_6,SPR_MUT_W4_7,SPR_MUT_W4_8,

    SPR_MUT_PAIN_1,SPR_MUT_DIE_1,SPR_MUT_DIE_2,SPR_MUT_DIE_3,
    SPR_MUT_PAIN_2,SPR_MUT_DEAD,

    SPR_MUT_SHOOT1,SPR_MUT_SHOOT2,SPR_MUT_SHOOT3,SPR_MUT_SHOOT4,
    SPR_MUT_SHOOT5,

//
// officer
//
    SPR_OFC_S_1,SPR_OFC_S_2,SPR_OFC_S_3,SPR_OFC_S_4,
    SPR_OFC_S_5,SPR_OFC_S_6,SPR_OFC_S_7,SPR_OFC_S_8,

    SPR_OFC_W1_1,SPR_OFC_W1_2,SPR_OFC_W1_3,SPR_OFC_W1_4,
    SPR_OFC_W1_5,SPR_OFC_W1_6,SPR_OFC_W1_7,SPR_OFC_W1_8,

    SPR_OFC_W2_1,SPR_OFC_W2_2,SPR_OFC_W2_3,SPR_OFC_W2_4,
    SPR_OFC_W2_5,SPR_OFC_W2_6,SPR_OFC_W2_7,SPR_OFC_W2_8,

    SPR_OFC_W3_1,SPR_OFC_W3_2,SPR_OFC_W3_3,SPR_OFC_W3_4,
    SPR_OFC_W3_5,SPR_OFC_W3_6,SPR_OFC_W3_7,SPR_OFC_W3_8,

    SPR_OFC_W4_1,SPR_OFC_W4_2,SPR_OFC_W4_3,SPR_OFC_W4_4,
    SPR_OFC_W4_5,SPR_OFC_W4_6,SPR_OFC_W4_7,SPR_OFC_W4_8,

    SPR_OFC_PAIN_1,SPR_OFC_DIE_1,SPR_OFC_DIE_2,SPR_OFC_DIE_3,
    SPR_OFC_PAIN_2,SPR_OFC_DEAD,

    SPR_OFC_SHOOT1,SPR_OFC_SHOOT2,SPR_OFC_SHOOT3,

    SPR_STAT_46,

//
// ghosts
//
    SPR_SNIPER_PAIN_1,SPR_SNIPER_DIE_1,SPR_SNIPER_DIE_2,SPR_SNIPER_DIE_3,
    SPR_SNIPER_PAIN_2,SPR_SNIPER_DEAD,

    SPR_SNIPER_SHOOT1,SPR_SNIPER_SHOOT2,SPR_SNIPER_SHOOT3,

    SPR_STAT_52_1,SPR_STAT_52_2,SPR_STAT_53,SPR_STRANGLEDGUARD,
    SPR_STAT_54_1,SPR_STAT_54_2,

//
// hans
//
    SPR_BOSS_W1,SPR_BOSS_W2,SPR_BOSS_W3,SPR_BOSS_W4,
    SPR_BOSS_SHOOT1,SPR_BOSS_SHOOT2,SPR_BOSS_SHOOT3,SPR_BOSS_DEAD,

    SPR_BOSS_DIE1,SPR_BOSS_DIE2,SPR_BOSS_DIE3,

//
// schabbs
//
    SPR_SCHABB_W1,SPR_SCHABB_W2,SPR_SCHABB_W3,SPR_SCHABB_W4,
    SPR_SCHABB_SHOOT1,SPR_SCHABB_SHOOT2,SPR_SCHABB_SHOOT3,SPR_SCHABB_DEAD,

    SPR_SCHABB_DIE1,SPR_SCHABB_DIE2,SPR_SCHABB_DIE3,

    SPR_HYPO1,SPR_HYPO2,SPR_HYPO3,SPR_HYPO4,

//
// fake
//
    SPR_FAKE_W1,SPR_FAKE_W2,SPR_FAKE_W3,SPR_FAKE_W4,
    SPR_FAKE_SHOOT1,SPR_FAKE_SHOOT2,SPR_FAKE_SHOOT3,SPR_FAKE_DEAD,

    SPR_FAKE_DIE1,SPR_FAKE_DIE2,SPR_FAKE_DIE3,

    SPR_FIRE1,SPR_FIRE2,SPR_FIRE3,SPR_FIRE4,
    SPR_FIREHIT1,SPR_FIREHIT2,SPR_FIREHIT3,

//
// death mutant
//
    SPR_MECHA_LAUNCH,

    SPR_MECHA_W1,SPR_MECHA_W2,SPR_MECHA_W3,SPR_MECHA_W4,
    SPR_MECHA_SHOOT1,SPR_MECHA_SHOOT2,SPR_MECHA_SHOOT3,SPR_MECHA_DEAD,

    SPR_MECHA_DIE1,SPR_MECHA_DIE2,SPR_MECHA_DIE3,

    SPR_TOTEN_W1,SPR_TOTEN_F1,SPR_TOTEN_F2,SPR_TOTEN_F3,
    SPR_TOTEN_F4,

//
// giftmacher
//
    SPR_GIFT_W1,SPR_GIFT_W2,SPR_GIFT_W3,SPR_GIFT_W4,
    SPR_GIFT_SHOOT1,SPR_GIFT_SHOOT2,SPR_GIFT_SHOOT3,SPR_GIFT_DEAD,

    SPR_GIFT_DIE1,SPR_GIFT_DIE2,SPR_GIFT_DIE3,

//
// Rocket, smoke and small explosion
//
    SPR_ROCKET_1,SPR_ROCKET_2,SPR_ROCKET_3,SPR_ROCKET_4,
    SPR_ROCKET_5,SPR_ROCKET_6,SPR_ROCKET_7,SPR_ROCKET_8,

    SPR_SMOKE_1,SPR_SMOKE_2,SPR_SMOKE_3,SPR_SMOKE_4,
    SPR_BOOM_1,SPR_BOOM_2,SPR_BOOM_3,

    SPR_HROCKET_1,SPR_HROCKET_2,SPR_HROCKET_3,SPR_HROCKET_4,
    SPR_HROCKET_5,SPR_HROCKET_6,SPR_HROCKET_7,SPR_HROCKET_8,

    SPR_HSMOKE_1,SPR_HSMOKE_2,SPR_HSMOKE_3,SPR_HSMOKE_4,
    SPR_HBOOM_1,SPR_HBOOM_2,SPR_HBOOM_3,

    SPR_EROCKET_1,SPR_EROCKET_2,SPR_EROCKET_3,SPR_EROCKET_4,
    SPR_EROCKET_5,SPR_EROCKET_6,SPR_EROCKET_7,SPR_EROCKET_8,

    SPR_ESMOKE_1,SPR_ESMOKE_2,SPR_ESMOKE_3,SPR_ESMOKE_4,
    SPR_EBOOM_1,SPR_EBOOM_2,SPR_EBOOM_3,

    SPR_FLAKBOOM_1,SPR_FLAKBOOM_2,SPR_FLAKBOOM_3,

//
// gretel
//
    SPR_GRETEL_W1,SPR_GRETEL_W2,SPR_GRETEL_W3,SPR_GRETEL_W4,
    SPR_GRETEL_SHOOT1,SPR_GRETEL_SHOOT2,SPR_GRETEL_SHOOT3,SPR_GRETEL_DEAD,

    SPR_GRETEL_DIE1,SPR_GRETEL_DIE2,SPR_GRETEL_DIE3,

//
// fat face
//
    SPR_FAT_LAUNCH,

    SPR_FAT_W1,SPR_FAT_W2,SPR_FAT_W3,SPR_FAT_W4,
    SPR_FAT_SHOOT1,SPR_FAT_SHOOT2,SPR_FAT_SHOOT3,SPR_FAT_DEAD,

    SPR_FAT_DIE1,SPR_FAT_DIE2,SPR_FAT_DIE3,

//
// Panz Grosse
//
    SPR_PANZ_LAUNCH,

    SPR_PANZ_W1,SPR_PANZ_W2,SPR_PANZ_W3,SPR_PANZ_W4,
    SPR_PANZ_SHOOT1,SPR_PANZ_SHOOT2,SPR_PANZ_SHOOT3,SPR_PANZ_DEAD,

    SPR_PANZ_DIE1,SPR_PANZ_DIE2,SPR_PANZ_DIE3,

//
// new static objects
//
    SPR_STAT2_0,SPR_STAT2_1,SPR_STAT2_2,SPR_STAT2_3,
    SPR_STAT2_4,SPR_STAT2_5,SPR_STAT2_6,SPR_STAT2_7,

    SPR_STAT2_8,SPR_STAT2_9,SPR_STAT2_10,SPR_STAT2_11,
    SPR_STAT2_12,SPR_STAT2_13,SPR_STAT2_14,SPR_STAT2_15,

    SPR_STAT2_16,SPR_STAT2_17,SPR_STAT2_18,SPR_STAT2_19,
    SPR_STAT2_20,SPR_STAT2_21,SPR_STAT2_22,SPR_STAT2_23,

    SPR_STAT2_24,SPR_STAT2_25,SPR_STAT2_26,SPR_STAT2_27,
    SPR_STAT2_28,SPR_STAT2_29,SPR_STAT2_30,SPR_STAT2_31_1,

    SPR_STAT2_31_2,SPR_STAT2_32,SPR_STAT2_33_1,SPR_STAT2_33_2,
    SPR_STAT2_34,SPR_STAT2_35,SPR_STAT2_36,SPR_STAT2_37,

    SPR_STAT2_38,SPR_STAT2_39,SPR_STAT2_40,SPR_STAT2_41,
    SPR_STAT2_42,SPR_STAT2_43,

//
// guard 2
//
    SPR_GRD2_S_1,SPR_GRD2_S_2,SPR_GRD2_S_3,SPR_GRD2_S_4,
    SPR_GRD2_S_5,SPR_GRD2_S_6,SPR_GRD2_S_7,SPR_GRD2_S_8,

    SPR_GRD2_W1_1,SPR_GRD2_W1_2,SPR_GRD2_W1_3,SPR_GRD2_W1_4,
    SPR_GRD2_W1_5,SPR_GRD2_W1_6,SPR_GRD2_W1_7,SPR_GRD2_W1_8,

    SPR_GRD2_W2_1,SPR_GRD2_W2_2,SPR_GRD2_W2_3,SPR_GRD2_W2_4,
    SPR_GRD2_W2_5,SPR_GRD2_W2_6,SPR_GRD2_W2_7,SPR_GRD2_W2_8,

    SPR_GRD2_W3_1,SPR_GRD2_W3_2,SPR_GRD2_W3_3,SPR_GRD2_W3_4,
    SPR_GRD2_W3_5,SPR_GRD2_W3_6,SPR_GRD2_W3_7,SPR_GRD2_W3_8,

    SPR_GRD2_W4_1,SPR_GRD2_W4_2,SPR_GRD2_W4_3,SPR_GRD2_W4_4,
    SPR_GRD2_W4_5,SPR_GRD2_W4_6,SPR_GRD2_W4_7,SPR_GRD2_W4_8,

    SPR_GRD2_PAIN_1,SPR_GRD2_DIE_1,SPR_GRD2_DIE_2,SPR_GRD2_DIE_3,
    SPR_GRD2_PAIN_2,SPR_GRD2_DEAD,

    SPR_GRD2_SHOOT1,SPR_GRD2_SHOOT2,SPR_GRD2_SHOOT3,

//
// rifle guard
//
    SPR_RIFLE_S_1,SPR_RIFLE_S_2,SPR_RIFLE_S_3,SPR_RIFLE_S_4,
    SPR_RIFLE_S_5,SPR_RIFLE_S_6,SPR_RIFLE_S_7,SPR_RIFLE_S_8,

    SPR_RIFLE_W1_1,SPR_RIFLE_W1_2,SPR_RIFLE_W1_3,SPR_RIFLE_W1_4,
    SPR_RIFLE_W1_5,SPR_RIFLE_W1_6,SPR_RIFLE_W1_7,SPR_RIFLE_W1_8,

    SPR_RIFLE_W2_1,SPR_RIFLE_W2_2,SPR_RIFLE_W2_3,SPR_RIFLE_W2_4,
    SPR_RIFLE_W2_5,SPR_RIFLE_W2_6,SPR_RIFLE_W2_7,SPR_RIFLE_W2_8,

    SPR_RIFLE_W3_1,SPR_RIFLE_W3_2,SPR_RIFLE_W3_3,SPR_RIFLE_W3_4,
    SPR_RIFLE_W3_5,SPR_RIFLE_W3_6,SPR_RIFLE_W3_7,SPR_RIFLE_W3_8,

    SPR_RIFLE_W4_1,SPR_RIFLE_W4_2,SPR_RIFLE_W4_3,SPR_RIFLE_W4_4,
    SPR_RIFLE_W4_5,SPR_RIFLE_W4_6,SPR_RIFLE_W4_7,SPR_RIFLE_W4_8,

    SPR_RIFLE_PAIN_1,SPR_RIFLE_DIE_1,SPR_RIFLE_DIE_2,SPR_RIFLE_DIE_3,
    SPR_RIFLE_PAIN_2,SPR_RIFLE_DEAD,

    SPR_RIFLE_SHOOT1,SPR_RIFLE_SHOOT2,SPR_RIFLE_SHOOT3,

//
// mp40 guard
//
    SPR_INF_S_1,SPR_INF_S_2,SPR_INF_S_3,SPR_INF_S_4,
    SPR_INF_S_5,SPR_INF_S_6,SPR_INF_S_7,SPR_INF_S_8,

    SPR_INF_W1_1,SPR_INF_W1_2,SPR_INF_W1_3,SPR_INF_W1_4,
    SPR_INF_W1_5,SPR_INF_W1_6,SPR_INF_W1_7,SPR_INF_W1_8,

    SPR_INF_W2_1,SPR_INF_W2_2,SPR_INF_W2_3,SPR_INF_W2_4,
    SPR_INF_W2_5,SPR_INF_W2_6,SPR_INF_W2_7,SPR_INF_W2_8,

    SPR_INF_W3_1,SPR_INF_W3_2,SPR_INF_W3_3,SPR_INF_W3_4,
    SPR_INF_W3_5,SPR_INF_W3_6,SPR_INF_W3_7,SPR_INF_W3_8,

    SPR_INF_W4_1,SPR_INF_W4_2,SPR_INF_W4_3,SPR_INF_W4_4,
    SPR_INF_W4_5,SPR_INF_W4_6,SPR_INF_W4_7,SPR_INF_W4_8,

    SPR_INF_PAIN_1,SPR_INF_DIE_1,SPR_INF_DIE_2,SPR_INF_DIE_3,
    SPR_INF_PAIN_2,SPR_INF_DEAD,

    SPR_INF_SHOOT1,SPR_INF_SHOOT2,SPR_INF_SHOOT3,

//
// mp40 guard 2
//
    SPR_INF2_S_1,SPR_INF2_S_2,SPR_INF2_S_3,SPR_INF2_S_4,
    SPR_INF2_S_5,SPR_INF2_S_6,SPR_INF2_S_7,SPR_INF2_S_8,

    SPR_INF2_W1_1,SPR_INF2_W1_2,SPR_INF2_W1_3,SPR_INF2_W1_4,
    SPR_INF2_W1_5,SPR_INF2_W1_6,SPR_INF2_W1_7,SPR_INF2_W1_8,

    SPR_INF2_W2_1,SPR_INF2_W2_2,SPR_INF2_W2_3,SPR_INF2_W2_4,
    SPR_INF2_W2_5,SPR_INF2_W2_6,SPR_INF2_W2_7,SPR_INF2_W2_8,

    SPR_INF2_W3_1,SPR_INF2_W3_2,SPR_INF2_W3_3,SPR_INF2_W3_4,
    SPR_INF2_W3_5,SPR_INF2_W3_6,SPR_INF2_W3_7,SPR_INF2_W3_8,

    SPR_INF2_W4_1,SPR_INF2_W4_2,SPR_INF2_W4_3,SPR_INF2_W4_4,
    SPR_INF2_W4_5,SPR_INF2_W4_6,SPR_INF2_W4_7,SPR_INF2_W4_8,

    SPR_INF2_PAIN_1,SPR_INF2_DIE_1,SPR_INF2_DIE_2,SPR_INF2_DIE_3,
    SPR_INF2_PAIN_2,SPR_INF2_DEAD,

    SPR_INF2_SHOOT1,SPR_INF2_SHOOT2,SPR_INF2_SHOOT3,

//
// mp40 officer
//
    SPR_OFC2_S_1,SPR_OFC2_S_2,SPR_OFC2_S_3,SPR_OFC2_S_4,
    SPR_OFC2_S_5,SPR_OFC2_S_6,SPR_OFC2_S_7,SPR_OFC2_S_8,

    SPR_OFC2_W1_1,SPR_OFC2_W1_2,SPR_OFC2_W1_3,SPR_OFC2_W1_4,
    SPR_OFC2_W1_5,SPR_OFC2_W1_6,SPR_OFC2_W1_7,SPR_OFC2_W1_8,

    SPR_OFC2_W2_1,SPR_OFC2_W2_2,SPR_OFC2_W2_3,SPR_OFC2_W2_4,
    SPR_OFC2_W2_5,SPR_OFC2_W2_6,SPR_OFC2_W2_7,SPR_OFC2_W2_8,

    SPR_OFC2_W3_1,SPR_OFC2_W3_2,SPR_OFC2_W3_3,SPR_OFC2_W3_4,
    SPR_OFC2_W3_5,SPR_OFC2_W3_6,SPR_OFC2_W3_7,SPR_OFC2_W3_8,

    SPR_OFC2_W4_1,SPR_OFC2_W4_2,SPR_OFC2_W4_3,SPR_OFC2_W4_4,
    SPR_OFC2_W4_5,SPR_OFC2_W4_6,SPR_OFC2_W4_7,SPR_OFC2_W4_8,

    SPR_OFC2_PAIN_1,SPR_OFC2_DIE_1,SPR_OFC2_DIE_2,SPR_OFC2_DIE_3,
    SPR_OFC2_PAIN_2,SPR_OFC2_DEAD,

    SPR_OFC2_SHOOT1,SPR_OFC2_SHOOT2,SPR_OFC2_SHOOT3,

//
// mg42 gunner
//
    SPR_GRD_MG42_PAIN_1,SPR_GRD_MG42_DIE_1,SPR_GRD_MG42_DIE_2,SPR_GRD_MG42_DIE_3,
    SPR_GRD_MG42_PAIN_2,SPR_GRD_MG42_DEAD,

    SPR_GRD_MG42_SHOOT1,SPR_GRD_MG42_SHOOT2,SPR_GRD_MG42_SHOOT3,

//
// dog 2
//
    SPR_DOG2_W1_1,SPR_DOG2_W1_2,SPR_DOG2_W1_3,SPR_DOG2_W1_4,
    SPR_DOG2_W1_5,SPR_DOG2_W1_6,SPR_DOG2_W1_7,SPR_DOG2_W1_8,

    SPR_DOG2_W2_1,SPR_DOG2_W2_2,SPR_DOG2_W2_3,SPR_DOG2_W2_4,
    SPR_DOG2_W2_5,SPR_DOG2_W2_6,SPR_DOG2_W2_7,SPR_DOG2_W2_8,

    SPR_DOG2_W3_1,SPR_DOG2_W3_2,SPR_DOG2_W3_3,SPR_DOG2_W3_4,
    SPR_DOG2_W3_5,SPR_DOG2_W3_6,SPR_DOG2_W3_7,SPR_DOG2_W3_8,

    SPR_DOG2_W4_1,SPR_DOG2_W4_2,SPR_DOG2_W4_3,SPR_DOG2_W4_4,
    SPR_DOG2_W4_5,SPR_DOG2_W4_6,SPR_DOG2_W4_7,SPR_DOG2_W4_8,

    SPR_DOG2_DIE_1,SPR_DOG2_DIE_2,SPR_DOG2_DIE_3,SPR_DOG2_DEAD,
    SPR_DOG2_JUMP1,SPR_DOG2_JUMP2,SPR_DOG2_JUMP3,

//
// ss rifleman
//
    SPR_SSRIFLE_S_1,SPR_SSRIFLE_S_2,SPR_SSRIFLE_S_3,SPR_SSRIFLE_S_4,
    SPR_SSRIFLE_S_5,SPR_SSRIFLE_S_6,SPR_SSRIFLE_S_7,SPR_SSRIFLE_S_8,

    SPR_SSRIFLE_W1_1,SPR_SSRIFLE_W1_2,SPR_SSRIFLE_W1_3,SPR_SSRIFLE_W1_4,
    SPR_SSRIFLE_W1_5,SPR_SSRIFLE_W1_6,SPR_SSRIFLE_W1_7,SPR_SSRIFLE_W1_8,

    SPR_SSRIFLE_W2_1,SPR_SSRIFLE_W2_2,SPR_SSRIFLE_W2_3,SPR_SSRIFLE_W2_4,
    SPR_SSRIFLE_W2_5,SPR_SSRIFLE_W2_6,SPR_SSRIFLE_W2_7,SPR_SSRIFLE_W2_8,

    SPR_SSRIFLE_W3_1,SPR_SSRIFLE_W3_2,SPR_SSRIFLE_W3_3,SPR_SSRIFLE_W3_4,
    SPR_SSRIFLE_W3_5,SPR_SSRIFLE_W3_6,SPR_SSRIFLE_W3_7,SPR_SSRIFLE_W3_8,

    SPR_SSRIFLE_W4_1,SPR_SSRIFLE_W4_2,SPR_SSRIFLE_W4_3,SPR_SSRIFLE_W4_4,
    SPR_SSRIFLE_W4_5,SPR_SSRIFLE_W4_6,SPR_SSRIFLE_W4_7,SPR_SSRIFLE_W4_8,

    SPR_SSRIFLE_PAIN_1,SPR_SSRIFLE_DIE_1,SPR_SSRIFLE_DIE_2,SPR_SSRIFLE_DIE_3,
    SPR_SSRIFLE_PAIN_2,SPR_SSRIFLE_DEAD,

    SPR_SSRIFLE_SHOOT1,SPR_SSRIFLE_SHOOT2,SPR_SSRIFLE_SHOOT3,

//
// ss officer
//
    SPR_SSOFC_S_1,SPR_SSOFC_S_2,SPR_SSOFC_S_3,SPR_SSOFC_S_4,
    SPR_SSOFC_S_5,SPR_SSOFC_S_6,SPR_SSOFC_S_7,SPR_SSOFC_S_8,

    SPR_SSOFC_W1_1,SPR_SSOFC_W1_2,SPR_SSOFC_W1_3,SPR_SSOFC_W1_4,
    SPR_SSOFC_W1_5,SPR_SSOFC_W1_6,SPR_SSOFC_W1_7,SPR_SSOFC_W1_8,

    SPR_SSOFC_W2_1,SPR_SSOFC_W2_2,SPR_SSOFC_W2_3,SPR_SSOFC_W2_4,
    SPR_SSOFC_W2_5,SPR_SSOFC_W2_6,SPR_SSOFC_W2_7,SPR_SSOFC_W2_8,

    SPR_SSOFC_W3_1,SPR_SSOFC_W3_2,SPR_SSOFC_W3_3,SPR_SSOFC_W3_4,
    SPR_SSOFC_W3_5,SPR_SSOFC_W3_6,SPR_SSOFC_W3_7,SPR_SSOFC_W3_8,

    SPR_SSOFC_W4_1,SPR_SSOFC_W4_2,SPR_SSOFC_W4_3,SPR_SSOFC_W4_4,
    SPR_SSOFC_W4_5,SPR_SSOFC_W4_6,SPR_SSOFC_W4_7,SPR_SSOFC_W4_8,

    SPR_SSOFC_PAIN_1,SPR_SSOFC_DIE_1,SPR_SSOFC_DIE_2,SPR_SSOFC_DIE_3,
    SPR_SSOFC_PAIN_2,SPR_SSOFC_DEAD,

    SPR_SSOFC_SHOOT1,SPR_SSOFC_SHOOT2,SPR_SSOFC_SHOOT3,

//
// ss officer 2
//
    SPR_SSOFC2_S_1,SPR_SSOFC2_S_2,SPR_SSOFC2_S_3,SPR_SSOFC2_S_4,
    SPR_SSOFC2_S_5,SPR_SSOFC2_S_6,SPR_SSOFC2_S_7,SPR_SSOFC2_S_8,

    SPR_SSOFC2_W1_1,SPR_SSOFC2_W1_2,SPR_SSOFC2_W1_3,SPR_SSOFC2_W1_4,
    SPR_SSOFC2_W1_5,SPR_SSOFC2_W1_6,SPR_SSOFC2_W1_7,SPR_SSOFC2_W1_8,

    SPR_SSOFC2_W2_1,SPR_SSOFC2_W2_2,SPR_SSOFC2_W2_3,SPR_SSOFC2_W2_4,
    SPR_SSOFC2_W2_5,SPR_SSOFC2_W2_6,SPR_SSOFC2_W2_7,SPR_SSOFC2_W2_8,

    SPR_SSOFC2_W3_1,SPR_SSOFC2_W3_2,SPR_SSOFC2_W3_3,SPR_SSOFC2_W3_4,
    SPR_SSOFC2_W3_5,SPR_SSOFC2_W3_6,SPR_SSOFC2_W3_7,SPR_SSOFC2_W3_8,

    SPR_SSOFC2_W4_1,SPR_SSOFC2_W4_2,SPR_SSOFC2_W4_3,SPR_SSOFC2_W4_4,
    SPR_SSOFC2_W4_5,SPR_SSOFC2_W4_6,SPR_SSOFC2_W4_7,SPR_SSOFC2_W4_8,

    SPR_SSOFC2_PAIN_1,SPR_SSOFC2_DIE_1,SPR_SSOFC2_DIE_2,SPR_SSOFC2_DIE_3,
    SPR_SSOFC2_PAIN_2,SPR_SSOFC2_DEAD,

    SPR_SSOFC2_SHOOT1,SPR_SSOFC2_SHOOT2,SPR_SSOFC2_SHOOT3,SPR_SSOFC2_SHOOT4,

//
// ss mg42 gunner
//
    SPR_SS_MG42_PAIN_1,SPR_SS_MG42_DIE_1,SPR_SS_MG42_DIE_2,SPR_SS_MG42_DIE_3,
    SPR_SS_MG42_PAIN_2,SPR_SS_MG42_DEAD,

    SPR_SS_MG42_SHOOT1,SPR_SS_MG42_SHOOT2,SPR_SS_MG42_SHOOT3,

//
// scientist
//
    SPR_SCI_S_1,SPR_SCI_S_2,SPR_SCI_S_3,SPR_SCI_S_4,
    SPR_SCI_S_5,SPR_SCI_S_6,SPR_SCI_S_7,SPR_SCI_S_8,

    SPR_SCI_W1_1,SPR_SCI_W1_2,SPR_SCI_W1_3,SPR_SCI_W1_4,
    SPR_SCI_W1_5,SPR_SCI_W1_6,SPR_SCI_W1_7,SPR_SCI_W1_8,

    SPR_SCI_W2_1,SPR_SCI_W2_2,SPR_SCI_W2_3,SPR_SCI_W2_4,
    SPR_SCI_W2_5,SPR_SCI_W2_6,SPR_SCI_W2_7,SPR_SCI_W2_8,

    SPR_SCI_W3_1,SPR_SCI_W3_2,SPR_SCI_W3_3,SPR_SCI_W3_4,
    SPR_SCI_W3_5,SPR_SCI_W3_6,SPR_SCI_W3_7,SPR_SCI_W3_8,

    SPR_SCI_W4_1,SPR_SCI_W4_2,SPR_SCI_W4_3,SPR_SCI_W4_4,
    SPR_SCI_W4_5,SPR_SCI_W4_6,SPR_SCI_W4_7,SPR_SCI_W4_8,

    SPR_SCI_PAIN_1,SPR_SCI_DIE_1,SPR_SCI_DIE_2,SPR_SCI_DIE_3,
    SPR_SCI_PAIN_2,SPR_SCI_DEAD,

    SPR_SCI_SHOOT1,SPR_SCI_SHOOT2,SPR_SCI_SHOOT3,

//
// scientist 2
//
    SPR_SCI2_S_1,SPR_SCI2_S_2,SPR_SCI2_S_3,SPR_SCI2_S_4,
    SPR_SCI2_S_5,SPR_SCI2_S_6,SPR_SCI2_S_7,SPR_SCI2_S_8,

    SPR_SCI2_W1_1,SPR_SCI2_W1_2,SPR_SCI2_W1_3,SPR_SCI2_W1_4,
    SPR_SCI2_W1_5,SPR_SCI2_W1_6,SPR_SCI2_W1_7,SPR_SCI2_W1_8,

    SPR_SCI2_W2_1,SPR_SCI2_W2_2,SPR_SCI2_W2_3,SPR_SCI2_W2_4,
    SPR_SCI2_W2_5,SPR_SCI2_W2_6,SPR_SCI2_W2_7,SPR_SCI2_W2_8,

    SPR_SCI2_W3_1,SPR_SCI2_W3_2,SPR_SCI2_W3_3,SPR_SCI2_W3_4,
    SPR_SCI2_W3_5,SPR_SCI2_W3_6,SPR_SCI2_W3_7,SPR_SCI2_W3_8,

    SPR_SCI2_W4_1,SPR_SCI2_W4_2,SPR_SCI2_W4_3,SPR_SCI2_W4_4,
    SPR_SCI2_W4_5,SPR_SCI2_W4_6,SPR_SCI2_W4_7,SPR_SCI2_W4_8,

    SPR_SCI2_PAIN_1,SPR_SCI2_DIE_1,SPR_SCI2_DIE_2,SPR_SCI2_DIE_3,
    SPR_SCI2_PAIN_2,SPR_SCI2_DEAD,

    SPR_SCI2_THROW1,SPR_SCI2_THROW2,SPR_SCI2_THROW3,

//
// waffen ss
//
    SPR_WAFFEN_S_1,SPR_WAFFEN_S_2,SPR_WAFFEN_S_3,SPR_WAFFEN_S_4,
    SPR_WAFFEN_S_5,SPR_WAFFEN_S_6,SPR_WAFFEN_S_7,SPR_WAFFEN_S_8,

    SPR_WAFFEN_W1_1,SPR_WAFFEN_W1_2,SPR_WAFFEN_W1_3,SPR_WAFFEN_W1_4,
    SPR_WAFFEN_W1_5,SPR_WAFFEN_W1_6,SPR_WAFFEN_W1_7,SPR_WAFFEN_W1_8,

    SPR_WAFFEN_W2_1,SPR_WAFFEN_W2_2,SPR_WAFFEN_W2_3,SPR_WAFFEN_W2_4,
    SPR_WAFFEN_W2_5,SPR_WAFFEN_W2_6,SPR_WAFFEN_W2_7,SPR_WAFFEN_W2_8,

    SPR_WAFFEN_W3_1,SPR_WAFFEN_W3_2,SPR_WAFFEN_W3_3,SPR_WAFFEN_W3_4,
    SPR_WAFFEN_W3_5,SPR_WAFFEN_W3_6,SPR_WAFFEN_W3_7,SPR_WAFFEN_W3_8,

    SPR_WAFFEN_W4_1,SPR_WAFFEN_W4_2,SPR_WAFFEN_W4_3,SPR_WAFFEN_W4_4,
    SPR_WAFFEN_W4_5,SPR_WAFFEN_W4_6,SPR_WAFFEN_W4_7,SPR_WAFFEN_W4_8,

    SPR_WAFFEN_PAIN_1,SPR_WAFFEN_DIE_1,SPR_WAFFEN_DIE_2,SPR_WAFFEN_DIE_3,
    SPR_WAFFEN_PAIN_2,SPR_WAFFEN_DEAD,

    SPR_WAFFEN_SHOOT1,SPR_WAFFEN_SHOOT2,SPR_WAFFEN_SHOOT3,

//
// waffen ss 2
//
    SPR_WAFFEN2_S_1,SPR_WAFFEN2_S_2,SPR_WAFFEN2_S_3,SPR_WAFFEN2_S_4,
    SPR_WAFFEN2_S_5,SPR_WAFFEN2_S_6,SPR_WAFFEN2_S_7,SPR_WAFFEN2_S_8,

    SPR_WAFFEN2_W1_1,SPR_WAFFEN2_W1_2,SPR_WAFFEN2_W1_3,SPR_WAFFEN2_W1_4,
    SPR_WAFFEN2_W1_5,SPR_WAFFEN2_W1_6,SPR_WAFFEN2_W1_7,SPR_WAFFEN2_W1_8,

    SPR_WAFFEN2_W2_1,SPR_WAFFEN2_W2_2,SPR_WAFFEN2_W2_3,SPR_WAFFEN2_W2_4,
    SPR_WAFFEN2_W2_5,SPR_WAFFEN2_W2_6,SPR_WAFFEN2_W2_7,SPR_WAFFEN2_W2_8,

    SPR_WAFFEN2_W3_1,SPR_WAFFEN2_W3_2,SPR_WAFFEN2_W3_3,SPR_WAFFEN2_W3_4,
    SPR_WAFFEN2_W3_5,SPR_WAFFEN2_W3_6,SPR_WAFFEN2_W3_7,SPR_WAFFEN2_W3_8,

    SPR_WAFFEN2_W4_1,SPR_WAFFEN2_W4_2,SPR_WAFFEN2_W4_3,SPR_WAFFEN2_W4_4,
    SPR_WAFFEN2_W4_5,SPR_WAFFEN2_W4_6,SPR_WAFFEN2_W4_7,SPR_WAFFEN2_W4_8,

    SPR_WAFFEN2_PAIN_1,SPR_WAFFEN2_DIE_1,SPR_WAFFEN2_DIE_2,SPR_WAFFEN2_DIE_3,
    SPR_WAFFEN2_PAIN_2,SPR_WAFFEN2_DEAD,

    SPR_WAFFEN2_SHOOT1,SPR_WAFFEN2_SHOOT2,SPR_WAFFEN2_SHOOT3,

//
// waffen ss w/ gasmask
//
    SPR_WAFFENGAS_S_1,SPR_WAFFENGAS_S_2,SPR_WAFFENGAS_S_3,SPR_WAFFENGAS_S_4,
    SPR_WAFFENGAS_S_5,SPR_WAFFENGAS_S_6,SPR_WAFFENGAS_S_7,SPR_WAFFENGAS_S_8,

    SPR_WAFFENGAS_W1_1,SPR_WAFFENGAS_W1_2,SPR_WAFFENGAS_W1_3,SPR_WAFFENGAS_W1_4,
    SPR_WAFFENGAS_W1_5,SPR_WAFFENGAS_W1_6,SPR_WAFFENGAS_W1_7,SPR_WAFFENGAS_W1_8,

    SPR_WAFFENGAS_W2_1,SPR_WAFFENGAS_W2_2,SPR_WAFFENGAS_W2_3,SPR_WAFFENGAS_W2_4,
    SPR_WAFFENGAS_W2_5,SPR_WAFFENGAS_W2_6,SPR_WAFFENGAS_W2_7,SPR_WAFFENGAS_W2_8,

    SPR_WAFFENGAS_W3_1,SPR_WAFFENGAS_W3_2,SPR_WAFFENGAS_W3_3,SPR_WAFFENGAS_W3_4,
    SPR_WAFFENGAS_W3_5,SPR_WAFFENGAS_W3_6,SPR_WAFFENGAS_W3_7,SPR_WAFFENGAS_W3_8,

    SPR_WAFFENGAS_W4_1,SPR_WAFFENGAS_W4_2,SPR_WAFFENGAS_W4_3,SPR_WAFFENGAS_W4_4,
    SPR_WAFFENGAS_W4_5,SPR_WAFFENGAS_W4_6,SPR_WAFFENGAS_W4_7,SPR_WAFFENGAS_W4_8,

    SPR_WAFFENGAS_PAIN_1,SPR_WAFFENGAS_DIE_1,SPR_WAFFENGAS_DIE_2,SPR_WAFFENGAS_DIE_3,
    SPR_WAFFENGAS_PAIN_2,SPR_WAFFENGAS_DEAD,

    SPR_WAFFENGAS_SHOOT1,SPR_WAFFENGAS_SHOOT2,SPR_WAFFENGAS_SHOOT3,

//
// waffen ss officer
//
    SPR_WAFFENOFC_S_1,SPR_WAFFENOFC_S_2,SPR_WAFFENOFC_S_3,SPR_WAFFENOFC_S_4,
    SPR_WAFFENOFC_S_5,SPR_WAFFENOFC_S_6,SPR_WAFFENOFC_S_7,SPR_WAFFENOFC_S_8,

    SPR_WAFFENOFC_W1_1,SPR_WAFFENOFC_W1_2,SPR_WAFFENOFC_W1_3,SPR_WAFFENOFC_W1_4,
    SPR_WAFFENOFC_W1_5,SPR_WAFFENOFC_W1_6,SPR_WAFFENOFC_W1_7,SPR_WAFFENOFC_W1_8,

    SPR_WAFFENOFC_W2_1,SPR_WAFFENOFC_W2_2,SPR_WAFFENOFC_W2_3,SPR_WAFFENOFC_W2_4,
    SPR_WAFFENOFC_W2_5,SPR_WAFFENOFC_W2_6,SPR_WAFFENOFC_W2_7,SPR_WAFFENOFC_W2_8,

    SPR_WAFFENOFC_W3_1,SPR_WAFFENOFC_W3_2,SPR_WAFFENOFC_W3_3,SPR_WAFFENOFC_W3_4,
    SPR_WAFFENOFC_W3_5,SPR_WAFFENOFC_W3_6,SPR_WAFFENOFC_W3_7,SPR_WAFFENOFC_W3_8,

    SPR_WAFFENOFC_W4_1,SPR_WAFFENOFC_W4_2,SPR_WAFFENOFC_W4_3,SPR_WAFFENOFC_W4_4,
    SPR_WAFFENOFC_W4_5,SPR_WAFFENOFC_W4_6,SPR_WAFFENOFC_W4_7,SPR_WAFFENOFC_W4_8,

    SPR_WAFFENOFC_PAIN_1,SPR_WAFFENOFC_DIE_1,SPR_WAFFENOFC_DIE_2,SPR_WAFFENOFC_DIE_3,
    SPR_WAFFENOFC_PAIN_2,SPR_WAFFENOFC_DEAD,

    SPR_WAFFENOFC_SHOOT1,SPR_WAFFENOFC_SHOOT2,SPR_WAFFENOFC_SHOOT3,

//
// waffen ss mg42 gunner
//
    SPR_WAFFEN_MG42_PAIN_1,SPR_WAFFEN_MG42_DIE_1,SPR_WAFFEN_MG42_DIE_2,SPR_WAFFEN_MG42_DIE_3,
    SPR_WAFFEN_MG42_PAIN_2,SPR_WAFFEN_MG42_DEAD,

    SPR_WAFFEN_MG42_SHOOT1,SPR_WAFFEN_MG42_SHOOT2,SPR_WAFFEN_MG42_SHOOT3,

//
// mutant 2
//
    SPR_MUT2_S_1,SPR_MUT2_S_2,SPR_MUT2_S_3,SPR_MUT2_S_4,
    SPR_MUT2_S_5,SPR_MUT2_S_6,SPR_MUT2_S_7,SPR_MUT2_S_8,

    SPR_MUT2_W1_1,SPR_MUT2_W1_2,SPR_MUT2_W1_3,SPR_MUT2_W1_4,
    SPR_MUT2_W1_5,SPR_MUT2_W1_6,SPR_MUT2_W1_7,SPR_MUT2_W1_8,

    SPR_MUT2_W2_1,SPR_MUT2_W2_2,SPR_MUT2_W2_3,SPR_MUT2_W2_4,
    SPR_MUT2_W2_5,SPR_MUT2_W2_6,SPR_MUT2_W2_7,SPR_MUT2_W2_8,

    SPR_MUT2_W3_1,SPR_MUT2_W3_2,SPR_MUT2_W3_3,SPR_MUT2_W3_4,
    SPR_MUT2_W3_5,SPR_MUT2_W3_6,SPR_MUT2_W3_7,SPR_MUT2_W3_8,

    SPR_MUT2_W4_1,SPR_MUT2_W4_2,SPR_MUT2_W4_3,SPR_MUT2_W4_4,
    SPR_MUT2_W4_5,SPR_MUT2_W4_6,SPR_MUT2_W4_7,SPR_MUT2_W4_8,

    SPR_MUT2_PAIN_1,SPR_MUT2_DIE_1,SPR_MUT2_DIE_2,SPR_MUT2_DIE_3,
    SPR_MUT2_PAIN_2,SPR_MUT2_DEAD,

    SPR_MUT2_SHOOT1,SPR_MUT2_SHOOT2,SPR_MUT2_SHOOT3,SPR_MUT2_SHOOT4,

//
// Mechanoid Stormtrooper
//
    SPR_MST_DORMANT,

    SPR_MST_W1_1,SPR_MST_W1_2,SPR_MST_W1_3,SPR_MST_W1_4,
    SPR_MST_W1_5,SPR_MST_W1_6,SPR_MST_W1_7,SPR_MST_W1_8,

    SPR_MST_W2_1,SPR_MST_W2_2,SPR_MST_W2_3,SPR_MST_W2_4,
    SPR_MST_W2_5,SPR_MST_W2_6,SPR_MST_W2_7,SPR_MST_W2_8,

    SPR_MST_W3_1,SPR_MST_W3_2,SPR_MST_W3_3,SPR_MST_W3_4,
    SPR_MST_W3_5,SPR_MST_W3_6,SPR_MST_W3_7,SPR_MST_W3_8,

    SPR_MST_W4_1,SPR_MST_W4_2,SPR_MST_W4_3,SPR_MST_W4_4,
    SPR_MST_W4_5,SPR_MST_W4_6,SPR_MST_W4_7,SPR_MST_W4_8,

    SPR_MST_DIE_1,SPR_MST_DIE_2,SPR_MST_DIE_3,SPR_MST_DIE_4,
    SPR_MST_DEAD,SPR_MST_SHOOT1,SPR_MST_SHOOT2,SPR_MST_SHOOT3,

//
// special death frames
//
    SPR_GRD_GIB_1,SPR_GRD_GIB_2,SPR_GRD_GIB_3,SPR_GRD_GIB_4,
    SPR_GRD_GIB_5,SPR_GRD_GIB_6,SPR_GRD_GIB_7,SPR_GRD_GIB_8,
    SPR_GRD_GIBS,

    SPR_SS_GIB_1,SPR_SS_GIB_2,SPR_SS_GIB_3,SPR_SS_GIB_4,
    SPR_SS_GIB_5,SPR_SS_GIB_6,SPR_SS_GIB_7,SPR_SS_GIB_8,
    SPR_SS_GIBS,

    SPR_MUT_GIB_1,SPR_MUT_GIB_2,SPR_MUT_GIB_3,SPR_MUT_GIB_4,
    SPR_MUT_GIB_5,SPR_MUT_GIB_6,SPR_MUT_GIB_7,SPR_MUT_GIB_8,
    SPR_MUT_GIBS,

    SPR_SCI_GIB_1,SPR_SCI_GIB_2,SPR_SCI_GIB_3,SPR_SCI_GIB_4,
    SPR_SCI_GIB_5,SPR_SCI_GIB_6,SPR_SCI_GIB_7,SPR_SCI_GIB_8,
    SPR_SCI_GIBS,

    SPR_WAFFEN_GIB_1,SPR_WAFFEN_GIB_2,SPR_WAFFEN_GIB_3,SPR_WAFFEN_GIB_4,
    SPR_WAFFEN_GIB_5,SPR_WAFFEN_GIB_6,SPR_WAFFEN_GIB_7,SPR_WAFFEN_GIB_8,
    SPR_WAFFEN_GIBS,

    SPR_GRD_BURN_1,SPR_GRD_BURN_2,SPR_GRD_BURN_3,SPR_GRD_BURN_4,
    SPR_GRD_BURN_5,SPR_GRD_BURN_6,SPR_GRD_BURN_7,SPR_GRD_BURNED,

//
// bullet effects
//
    SPR_TRACER_1,SPR_TRACER_2,SPR_TRACER_3,SPR_TRACER_4,
    SPR_TRACER_5,SPR_TRACER_6,SPR_TRACER_7,SPR_TRACER_8,
    SPR_RICOCHET_1,SPR_RICOCHET_2,SPR_RICOCHET_3,SPR_RICOCHET_4,

    SPR_HTRACER_1,SPR_HTRACER_2,SPR_HTRACER_3,SPR_HTRACER_4,
    SPR_HTRACER_5,SPR_HTRACER_6,SPR_HTRACER_7,SPR_HTRACER_8,
    SPR_HRICOCHET_1,SPR_HRICOCHET_2,SPR_HRICOCHET_3,SPR_HRICOCHET_4,

    SPR_MTRACER_1,SPR_MTRACER_2,SPR_MTRACER_3,SPR_MTRACER_4,
    SPR_MTRACER_5,SPR_MTRACER_6,SPR_MTRACER_7,SPR_MTRACER_8,
    SPR_MRICOCHET_1,SPR_MRICOCHET_2,SPR_MRICOCHET_3,SPR_MRICOCHET_4,

    SPR_ETRACER_1,SPR_ETRACER_2,SPR_ETRACER_3,SPR_ETRACER_4,
    SPR_ETRACER_5,SPR_ETRACER_6,SPR_ETRACER_7,SPR_ETRACER_8,
    SPR_ERICOCHET_1,SPR_ERICOCHET_2,SPR_ERICOCHET_3,SPR_ERICOCHET_4,

    SPR_RBLOOD_1,SPR_RBLOOD_2,SPR_RBLOOD_3,SPR_RBLOOD_4,
    SPR_PBLOOD_1,SPR_PBLOOD_2,SPR_PBLOOD_3,SPR_PBLOOD_4,

    SPR_HSPARK_1,SPR_HSPARK_2,SPR_HSPARK_3,SPR_HSPARK_4,
    SPR_LSPARK_1,SPR_LSPARK_2,SPR_LSPARK_3,SPR_LSPARK_4,

//
// player attack frames
//
    SPR_KNIFEREADY,SPR_KNIFEATK1,SPR_KNIFEATK2,SPR_KNIFEATK3,
    SPR_KNIFEATK4,

    SPR_PISTOLREADY,SPR_PISTOLATK1,SPR_PISTOLATK2,SPR_PISTOLATK3,
    SPR_PISTOLATK4,

    SPR_DUALPISTOLREADY,SPR_DUALPISTOLATK1,SPR_DUALPISTOLATK2,SPR_DUALPISTOLATK3,
    SPR_DUALPISTOLATK4,SPR_DUALPISTOLATK5,SPR_DUALPISTOLATK6,

    SPR_MAUSERREADY,SPR_MAUSERATK1,SPR_MAUSERATK2,SPR_MAUSERATK3,
    SPR_MAUSERATK4,

    SPR_DUALMAUSERREADY,SPR_DUALMAUSERATK1,SPR_DUALMAUSERATK2,SPR_DUALMAUSERATK3,
    SPR_DUALMAUSERATK4,

    SPR_KAR98KREADY,SPR_KAR98KATK1,SPR_KAR98KATK2,SPR_KAR98KATK3,
    SPR_KAR98KATK4,SPR_KAR98KATK5,SPR_KAR98KATK6,SPR_KAR98KATK7,
    SPR_KAR98KATK8,SPR_KAR98KATK9,SPR_KAR98KATK10,SPR_KAR98KATK11,

    SPR_KAR98KSCOPE,
    SPR_SCOPEDKAR98KREADY,SPR_SCOPEDKAR98KATK1,SPR_SCOPEDKAR98KATK2,SPR_SCOPEDKAR98KATK3,
    SPR_SCOPEDKAR98KATK4,SPR_SCOPEDKAR98KATK5,SPR_SCOPEDKAR98KATK6,SPR_SCOPEDKAR98KATK7,
    SPR_SCOPEDKAR98KATK8,SPR_SCOPEDKAR98KATK9,SPR_SCOPEDKAR98KATK10,SPR_SCOPEDKAR98KATK11,

    SPR_MP40READY,SPR_MP40ATK1,SPR_MP40ATK2,MP40ATK3,
    SPR_MP40ATK4,

    SPR_DUALMP40READY,SPR_DUALMP40ATK1,SPR_DUALMP40ATK2,SPR_DUALMP40ATK3,
    SPR_DUALMP40ATK4,

    SPR_STG44READY,SPR_STG44ATK1,SPR_STG44ATK2,SPR_STG44ATK3,
    SPR_STG44ATK4,

    SPR_FLAMETHROWERREADY,SPR_FLAMETHROWERATK1,SPR_FLAMETHROWERATK2,SPR_FLAMETHROWERATK3,
    SPR_FLAMETHROWERATK4,

    SPR_MG42READY,SPR_MG42ATK1,SPR_MG42ATK2,SPR_MG42ATK3,
    SPR_MG42ATK4,

    SPR_CHAINGUNREADY,SPR_CHAINGUNATK1,SPR_CHAINGUNATK2,SPR_CHAINGUNATK3,
    SPR_CHAINGUNATK4,

    SPR_PANZERSCHRECKREADY,SPR_PANZERSCHRECKATK1,SPR_PANZERSCHRECKATK2,SPR_PANZERSCHRECKATK3,
    SPR_PANZERSCHRECKATK4,SPR_PANZERSCHRECKATK5,SPR_PANZERSCHRECKATK6,

    SPR_FLAKGUNREADY,SPR_FLAKGUN_ATK1,SPR_FLAKGUN_ATK2,SPR_FLAKGUN_ATK3,
    SPR_FLAKGUN_ATK4,

    SPR_MOUNTEDMG42READY,SPR_MOUNTEDMG42ATK1,SPR_MOUNTEDMG42ATK2,SPR_MOUNTEDMG42ATK3,
    SPR_MOUNTEDMG42ATK4,

};


/*
=============================================================================

                               GLOBAL TYPES

=============================================================================
*/

typedef enum {
    di_north,
    di_east,
    di_south,
    di_west
} controldir_t;

typedef enum {
    dr_normal,
    dr_lock1,
    dr_lock2,
    dr_lock3,
    dr_lock4,
    dr_elevator,
    dr_lab,
    dr_cell,
    dr_health,
    dr_ammo,
    dr_lock_fw,
    dr_lock_bw,
    dr_toten,
    NUMDOORTYPES
} door_t;

typedef enum {
    ac_badobject = -1,
    ac_no,
    ac_yes,
    ac_allways
} activetype;

typedef enum {
    nothing,
    playerobj,
    inertobj,
    guardobj,
    riflemanobj,
    infantryobj,
    officerobj,
    officer2obj,
    mg42guardobj,
    ssobj,
    ssriflemanobj,
    ssofficerobj,
    ssofficer2obj,
    mg42ssobj,
    sniperobj,
    dogobj,
    scientistobj,
    waffenssobj,
    waffenofficerobj,
    mg42waffenssobj,
    mutantobj,
    mechtrooperobj,
    flametrooperobj,
    chaingunnerobj,
    rockettrooperobj,
    bossobj,
    schabbobj,
    fatobj,
    panzobj,
    deatharmorobj,
    ghostobj,

    needleobj,
    fireobj,
    rocketobj,
    tracerobj,
    bjfireobj,
    bjrocketobj,
    sparkobj,
    NUMOBCLASSES
} classtype;

typedef enum {
    none,
    block,
    bo_pushwall,
    bo_acid,
    bo_alpo,
    bo_firstaid,
    bo_key1,
    bo_key2,
    bo_key3,
    bo_key4,
    bo_cross,
    bo_chalice,
    bo_bible,
    bo_crown,
    bo_clip,
    bo_clip_big,
    bo_pistol,
    bo_mauser,
    bo_mp40,
    bo_rounds,
    bo_rounds_big,
    bo_kar98k,
    bo_kar98k_sniper,
    bo_stg44ammo,
    bo_stg44,
    bo_fuel,
    bo_flamethrower,
    bo_ammobelt,
    bo_chaingun,
    bo_rocket,
    bo_panzerschreck,
    bo_flakammo,
    bo_flakgun_part,
    bo_mg42,
    bo_mg42_mounted,
    bo_armor_small,
    bo_armor_big,
    bo_armor_uber,
    bo_armor_power,
    bo_food,
    bo_fullheal,
    bo_regensphere,
    bo_spear
} wl_stat_t;

typedef enum {
    east,
    northeast,
    north,
    northwest,
    west,
    southwest,
    south,
    southeast,
    nodir
} dirtype;

enum {
    amb_none=-1,
    amb_default=0,
    amb_hall,
    amb_prison1,
    amb_prison2,
    amb_machine1,
    amb_machine2,
    amb_restroom,
    amb_dripping,
    amb_screams,
    amb_battle,
    amb_execution1,
    amb_execution2,
    NUMAMBIENCETYPES
};


typedef void (* statefunc) (void *);

typedef struct statestruct
{
    boolean rotate;
    short   shapenum;           // a shapenum of -1 means get from ob->temp1
    short   tictime;
    void    (*think) (void *),(*action) (void *);
    struct  statestruct *next;
    boolean fullbright;
} statetype;


//---------------------
//
// trivial actor structure
//
//---------------------

typedef struct statstruct
{
    byte      tilex,tiley;
    short     shapenum;           // if shapenum == -1 the obj has been removed
    byte      *visspot;
    uint32_t  flags;
    byte      itemnumber;
    byte      animframes;
} statobj_t;


//---------------------
//
// door actor structure
//
//---------------------

typedef enum
{
    dr_open,dr_closed,dr_opening,dr_closing
} doortype;

typedef struct doorstruct
{
    byte     tilex,tiley;
    boolean  vertical;
    byte     lock;
    doortype action;
    short    ticcount;
} doorobj_t;


//--------------------
//
// thinking actor structure
//
//--------------------

typedef enum
{
    alert_none,alert_sight,alert_hear,alert_dead,alert_findbody,alert_dormant
} alerttype;

typedef struct objstruct
{
    activetype  active;
    short       ticcount;
    classtype   obclass;
    statetype   *state;

    uint32_t    flags;              // FL_SHOOTABLE, etc

    int32_t     distance;           // if negative, wait for that door to open
    dirtype     dir;

    fixed       x,y;
    word        tilex,tiley;
    byte        areanumber;

    short       viewx;
    word        viewheight;
    fixed       transx,transy;      // in global coord

    short       angle;
    short       hitpoints;
    int32_t     speed;

    short       sproffset;
    short       ambientcount;
    alerttype   alert;

    short       temp1,temp2,hidden;
    struct objstruct *next,*prev;
} objtype;

enum
{
    bt_nobutton=-1,
    bt_attack=0,
    bt_strafe,
    bt_run,
    bt_use,
    bt_readyweapon1,
    bt_readyweapon2,
    bt_readyweapon3,
    bt_readyweapon4,
    bt_readyweapon5,
    bt_readyweapon6,
    bt_readyweapon7,
    bt_readyweapon8,
    bt_readyweapon9,
    bt_readyweapon10,
    bt_moveforward,
    bt_movebackward,
    bt_strafeleft,
    bt_straferight,
    bt_turnleft,
    bt_turnright,
    bt_nextweapon,
    bt_prevweapon,
    bt_esc,
    bt_pause,
    NUMBUTTONS
};


typedef enum
{
    wp_noweapon=-1,
    wp_knife,
    wp_pistol,
    wp_dualpistol,
    wp_mauser,
    wp_dualmauser,
    wp_kar98k,
    wp_sniper,
    wp_mp40,
    wp_dualmp40,
    wp_stg44,
    wp_flamethrower,
    wp_mg42,
    wp_chaingun,
    wp_panzerschreck,
    wp_flakgun,
    wp_mg42_mounted,
    NUMWEAPONS
} weapontype;


typedef enum
{
    ammo_none=-1,
    ammo_bullet,
    ammo_rifle,
    ammo_stg44,
    ammo_flame,
    ammo_chaingun,
    ammo_rocket,
    ammo_flak,
    NUMAMMOTYPES
} ammo_t;


typedef enum
{
    dam_none,
    dam_melee,
    dam_bullet,
    dam_mutagen,
    dam_fire,
    dam_machinegun,
    dam_chaingun,
    dam_flak,
    dam_rocket
} damage_t;


enum
{
    gd_baby,
    gd_easy,
    gd_medium,
    gd_hard,
    NUMDIFFICULTIES
};

//---------------
//
// gamestate structure
//
//---------------

typedef struct
{
    short       difficulty;
    short       mapon,nextmap;
    int32_t     oldscore,score,nextextra;
    short       lives;
    short       health,armor;
    short       ammo[NUMAMMOTYPES];
    short       ambience[NUMAREAS];    // WSJ: set during SetupGameLevel
    short       keys;
    weapontype  bestweapon,weapon,chosenweapon,lastchosenweapon[NUMBUTTONS];
    byte        hasweapon[NUMWEAPONS],flakgunparts;
    boolean     gotregenarmor,gotregenhealth;

    short       faceframe;
    short       attackframe,attackcount,weaponframe;

    short       episode,secretcount,treasurecount,killcount,
                secrettotal,treasuretotal,killtotal;
    int32_t     TimeCount,startbosshp;
    int32_t     killx,killy;
    boolean     victoryflag;            // set during victory animations
} gametype;


typedef enum
{
    ex_stillplaying,
    ex_completed,
    ex_died,
    ex_warped,
    ex_resetgame,
    ex_loadedgame,
    ex_victorious,
    ex_abort,
    ex_demodone,
    ex_secretlevel
} exit_t;


extern word *mapsegs[MAPPLANES];
extern int mapon;

/*
=============================================================================

                             WL_MAIN DEFINITIONS

=============================================================================
*/

extern bool nosound,allowwindow;

extern  boolean  loadedgame;
extern  fixed    focallength;
extern  int      viewscreenx, viewscreeny;
extern  int      viewwidth;
extern  int      viewheight;
extern  short    centerx;
extern  int32_t  heightnumerator;
extern  fixed    scale;

extern  int      dirangle[9];

extern  int      mouseadjustment;
extern  int      shootdelta;
extern  unsigned screenofs;

extern  boolean  startgame;
extern  char     str[80];
extern  char     configname[13];

//
// Command line parameter variables
//
extern  boolean  param_debugmode;
extern  boolean  param_nowait;
extern  int      param_difficulty;
extern  int      param_tedlevel;
extern  int      param_joystickindex;
extern  int      param_joystickhat;
extern  int      param_samplerate;
extern  int      param_audiobuffer;
extern  int      param_mission;
extern  boolean  param_goodtimes;
extern  boolean  param_ignorenumchunks;

#ifdef VIEASM
extern  boolean  param_8bitsound;
#endif


void            NewGame (int difficulty,int episode);
void            CalcProjection (int32_t focal);
void            NewViewSize (int width);
boolean         SetViewSize (unsigned width, unsigned height);
boolean         LoadTheGame(FILE *file,int x,int y);
boolean         SaveTheGame(FILE *file,int x,int y);
void            ShowViewSize (int width);
void            ShutdownId (void);


/*
=============================================================================

                         WL_GAME DEFINITIONS

=============================================================================
*/

extern  gametype        gamestate;
extern  byte            bordercol;
extern  byte            searchdata[0x10000];
extern  byte*           sdqstart;
extern  SDL_Surface     *latchpics[NUMLATCHPICS];

#ifdef DEBUGKEYS
extern  boolean         infinitelives;    // WSJ: used with tedlevel
#endif

void    ExitLevel (void);
void    PlaceValidArea (int tilex, int tiley);
void    SetupGameLevel (void);
void    GameLoop (void);
void    DrawPlayBorder (void);
void    DrawStatusBorder (byte color);
void    DrawPlayScreen (void);
void    DrawPlayBorderSides (void);
void    ShowActStatus();

void    PlayDemo (int demonumber);
void    RecordDemo (void);


#ifdef SPEAR
extern  int32_t            spearx,speary;
extern  unsigned        spearangle;
extern  boolean         spearflag;
#endif


#define ClearMemory SD_StopSound


// JAB
#define PlaySoundLocTile(s,c,tx,ty,f)    SD_PlaySound(s,c,TILECENTER(tx),TILECENTER(ty),f)
#define PlaySoundLocActor(s,c,ob,f)      SD_PlaySound(s,c,(ob)->x,(ob)->y,f)
#define PlaySoundGlobal(s,c,f)           SD_PlaySound(s,c,0,0,f)


/*
=============================================================================

                            WL_PLAY DEFINITIONS

=============================================================================
*/

#define BASEMOVE                35
#define RUNMOVE                 70
#define BASETURN                35
#define RUNTURN                 70

#define JOYSCALE                2

#define SNIPERMODE (gamestate.weapon==wp_sniper)

extern  byte            tilemap[MAPSIZE][MAPSIZE];      // wall values only
extern  byte            walltex[MAPSIZE][MAPSIZE];      // WSJ: wall texture values
extern  byte            spotvis[MAPSIZE][MAPSIZE];
extern  objtype         *actorat[MAPSIZE][MAPSIZE];

extern  objtype         *player;

extern  unsigned        tics;
extern  int             viewsize;

//
// current user input
//
extern  int         controlx,controly;              // range from -100 to 100
extern  boolean     buttonstate[NUMBUTTONS];
extern  objtype     objlist[MAXACTORS];
extern  boolean     buttonheld[NUMBUTTONS];
extern  exit_t      playstate;
extern  boolean     madenoise;
extern  byte        alertedarea,lastalertedarea;
extern  statobj_t   statobjlist[MAXSTATS];
extern  statobj_t   *laststatobj;
extern  objtype     *newobj,*killerobj;
extern  doorobj_t   doorobjlist[MAXDOORS];
extern  doorobj_t   *lastdoorobj;
extern  int         godmode,invisiblemode;

extern  boolean     demorecord,demoplayback;
extern  int8_t      *demoptr, *lastdemoptr;
extern  memptr      demobuffer;

//
// control info
//
extern  boolean     mouseenabled,mousevertmove,joystickenabled;
extern  int         dirscan[4];
extern  int         buttonscan[NUMBUTTONS];
extern  int         buttonmouse[4];
extern  int         buttonjoy[32];

void    InitActorList (void);
void    GetNewActor (void);
void    PlayLoop (void);

void    CenterWindow(word w,word h);

void    InitRedShifts (void);
void    FinishPaletteShifts (void);

void    RemoveObj (objtype *gone);
void    PollControls (void);
void    StartMusic(void);
void    StartDamageFlash (int damage);
void    StartBonusFlash (void);

extern  int32_t     curbosshp;

extern  objtype     *objfreelist;     // *obj,*player,*lastobj,

extern  boolean     noclip,ammocheat;
extern  int         singlestep, extravbls;

void    MakeNoise (objtype *ob);

/*
=============================================================================

                                WL_INTER

=============================================================================
*/

void IntroScreen (void);
void PG13(void);
void DrawHighScores(void);
void CheckHighScore (int32_t score,word other);
void Victory (void);
void LevelCompleted (void);
void ClearSplitVWB (void);

void PreloadGraphics(void);


/*
=============================================================================

                                WL_DEBUG

=============================================================================
*/

int DebugKeys (void);

/*
=============================================================================

                            WL_DRAW DEFINITIONS

=============================================================================
*/

//
// math tables
//
extern  short *pixelangle;
extern  int32_t finetangent[FINEANGLES/4];
extern  fixed sintable[];
extern  fixed *costable;
extern  int *wallheight;
extern  int32_t    lasttimecount;
extern  int32_t    frameon;

extern  unsigned screenloc[3];

extern  boolean fizzlein, fpscounter;

extern  fixed   viewx,viewy;                    // the focal point
extern  fixed   viewsin,viewcos;

void    ThreeDRefresh (void);
void    DrawScaleds (void);
void    CalcTics (void);

typedef struct
{
    word leftpix,rightpix;
    word dataofs[64];
// table data after dataofs[rightpix-leftpix+1]
} t_compshape;

extern  char    curmessage[50],cursubtitle[100];

void    GetMessage (const char *s);
void    GetSubtitle (const char *s);
void    PrintMessages (void);
boolean CheckVisTile (short tilex, short tiley);

/*
=============================================================================

                             WL_STATE DEFINITIONS

=============================================================================
*/
#define TURNTICS        10
#define SPDPATROL       512
#define SPDDOG          1500


void    InitHitRect (objtype *ob, unsigned radius);
void    SpawnNewObj (unsigned tilex, unsigned tiley, statetype *state);
void    NewState (objtype *ob, statetype *state);

boolean TryWalk (objtype *ob);
boolean SelectNextDir(objtype *ob, int dtilex, int dtiley);
void    SelectChaseDir (objtype *ob);
void    SelectDodgeDir (objtype *ob);
void    SelectRunDir (objtype *ob);
void    MoveObj (objtype *ob, int32_t move);
void    MoveObjInDir (objtype *ob, int32_t move, dirtype dir);
boolean SightPlayer (objtype *ob);

void    KillActor (objtype *ob, objtype *killer, damage_t damagetype);
boolean DamageActor (objtype *ob, objtype *attacker, unsigned damage, damage_t damagetype);

boolean CheckCoordLine (fixed fx1, fixed fy1, fixed fx2, fixed fy2);
#define CheckObjLine(ob,check)    CheckCoordLine(ob->x,ob->y,check->x,check->y)
boolean CheckTileLine (unsigned tx1, unsigned ty1, unsigned tx2, unsigned ty2);
boolean CheckLine (objtype *ob);
boolean CheckSight (objtype *ob);

/*
=============================================================================

                             WL_AGENT DEFINITIONS

=============================================================================
*/

extern  short    anglefrac;
extern  int      facecount, facetimes;
extern  word     plux,pluy;         // player coordinates scaled to unsigned
extern  int32_t  thrustspeed;
extern  objtype  *LastAttacker;
extern  short    maxammo[NUMAMMOTYPES];

const char keyname[4][20] = {
    "Gold Key",
    "Silver Key",
    "Bronze Key",
    "Elevator Key"
};

void    Thrust (int angle, int32_t speed);
void    StartInventory (void);
void    SpawnPlayer (int tilex, int tiley, int dir);
void    TakeDamage (int points, objtype *attacker, damage_t damagetype, boolean ignorearmor, boolean useflash);
void    GivePoints (int32_t points);
void    GetBonus (statobj_t *check);
boolean GiveWeapon (int weapon, int ammo);
boolean GiveAmmo (int ammo, ammo_t ammotype);
void    GiveKey (int key);
boolean GiveArmor (int points);
void    FireBullet (objtype *ob, short damage, damage_t damagetype, short spread, short numshots, short spriteoffset);

//
// player state info
//

void    StatusDrawFace(unsigned picnum);
void    DrawFace (void);
void    DrawHealth (void);
boolean HealSelf (int points);
void    DrawLevel (void);
void    DrawLives (void);
void    GiveExtraMan (void);
void    DrawScore (void);
void    DrawKeys (void);
void    DrawAmmo (void);


/*
=============================================================================

                             WL_ACT1 DEFINITIONS

=============================================================================
*/

extern  doorobj_t doorobjlist[MAXDOORS];
extern  doorobj_t *lastdoorobj;
extern  short     doornum;

extern  word      doorposition[MAXDOORS];

extern  byte      areaconnect[NUMAREAS][NUMAREAS];

extern  boolean   areabyplayer[NUMAREAS];

extern word     pwallstate;
extern word     pwallpos;        // amount a pushable wall has been moved (0-63)
extern word     pwallx,pwally;
extern byte     pwalldir,pwalltile;
extern byte     pwalltex;        // WSJ: texture on pushwall


void InitDoorList (void);
void InitStaticList (void);
void SpawnStatic (int tilex, int tiley, int type);
void SpawnSpecialStatic (int tilex, int tiley, int type, short picnum, uint32_t specialFlags, byte animframes);
#define SpawnSimpleStatic(x,y,picnum) SpawnSpecialStatic(x,y,none,picnum,0,0)
void SpawnDoor (int tilex, int tiley, boolean vertical, int lock);
void MoveDoors (void);
void MovePWalls (void);
void OpenDoor (int door);
void PlaceItemType (int itemtype, int tilex, int tiley);
boolean PushWall (int checkx, int checky, int dir);
boolean CheckLock (objtype *ob, int door);
void OperateDoor (int door);
void InitAreas (void);
void ConnectAreas (void);
boolean AreaByArea (int area1, int area2);
#define AreaByActor(ob,area)    AreaByArea(ob->areanumber,area)
#define AreaAt(x,y)             (MAPSPOT(x,y,0)-AREATILE)
#define VALIDAREATILE(t)        (t>=AREATILE && t<=LASTAREA)

/*
=============================================================================

                             WL_ACT2 DEFINITIONS

=============================================================================
*/

extern  statetype s_ricochet1;
extern  statetype s_bloodsplat1;

extern  statetype s_grddie1;
extern  statetype s_grdgib1;
extern  statetype s_grdburn1;
extern  statetype s_dogdie1;
extern  statetype s_mutburn1;
extern  statetype s_mstdie1;
extern  statetype s_bossdie1;
extern  statetype s_mechadie1;
extern  statetype s_narbdie1;
extern  statetype s_panzdie1;

extern  statetype s_grdchase1;
extern  statetype s_dogchase1;
extern  statetype s_mstchase1;
extern  statetype s_bosschase1;
extern  statetype s_bossshoot1;
extern  statetype s_schabbshoot1;

extern  statetype s_sniperwait;

extern  statetype s_grdpain;
extern  statetype s_grdpain1;
extern  statetype s_sniperpain;
extern  statetype s_sniperpain1;

extern  statetype s_deathcam;

extern  statetype s_schabbdeathcam2;
extern  statetype s_hitlerdeathcam2;
extern  statetype s_giftdeathcam2;
extern  statetype s_fatdeathcam2;

extern  statetype s_fire1;
extern  statetype s_rocket;

void SpawnStand (int which, int tilex, int tiley, int dir, short spriteoffset);
void SpawnPatrol (int which, int tilex, int tiley, int dir, short spriteoffset);
void SpawnElectrode (int tilex, int tiley);
boolean CheckBoss (objtype *ob);

void A_AlertSound (objtype *ob);
void A_DeathScream (objtype *ob);
void SpawnBulletPuff (fixed puffx, fixed puffy, statetype *state, short spriteoffset);
void SpawnBloodPuff (objtype *ob, fixed puffx, fixed puffy);
void SpawnRicochet (fixed puffx, fixed puffy, short spriteoffset, damage_t damagetype);
void SpawnTracer (fixed tracerx, fixed tracery, short angle, short spriteoffset);
void SpawnFlakPuff (objtype *ob, fixed puffx, fixed puffy, short damage, short spriteoffset);
void ProjFire (objtype *ob, classtype type, statetype *state, int32_t speed, short spriteoffset);
void ProjOffsFire (objtype *ob, classtype type, statetype *state, int32_t speed, short spriteoffset, short angleoffs);

void ChatterSound (objtype *ob, short tics);
boolean GuardChatterSound (objtype *ob);

/*
=============================================================================

                             WL_TEXT DEFINITIONS

=============================================================================
*/

extern  char    helpfilename[],endfilename[];

extern  void    HelpScreens(void);
extern  void    EndText(void);
extern  void    IntermissionScreens(void);


/*
=============================================================================

                            DREAMCAST DEFINITIONS

=============================================================================
*/

#ifdef _arch_dreamcast

// defined in dc_main.cpp
void DC_Main(void);
void DC_CheckParameters(void);
int DC_MousePresent(void);

// defined in dc_vmu.cpp
void StatusDrawLCD(int lcd_index);
int DC_SaveToVMU(char *src, int tp);
int DC_LoadFromVMU(char *dst);

#else

// ignore calls to this function
#define StatusDrawLCD(x)

#endif


/*
=============================================================================

                               GP2X DEFINITIONS

=============================================================================
*/

#ifdef GP2X

void GP2X_Init();
void GP2X_Shutdown();
void GP2X_ButtonDown(int button);
void GP2X_ButtonUp(int button);

#endif


/*
=============================================================================

                             MISC DEFINITIONS

=============================================================================
*/

static inline fixed FixedMul(fixed a, fixed b)
{
	return (fixed)(((int64_t)a * b + 0x8000) >> 16);
}

#ifdef PLAYDEMOLIKEORIGINAL
    #define DEMOCHOOSE_ORIG_SDL(orig, sdl) ((demorecord || demoplayback) ? (orig) : (sdl))
    #define DEMOCOND_ORIG                  (demorecord || demoplayback)
    #define DEMOIF_SDL                     if(DEMOCOND_SDL)
#else
    #define DEMOCHOOSE_ORIG_SDL(orig, sdl) (sdl)
    #define DEMOCOND_ORIG                  false
    #define DEMOIF_SDL
#endif
#define DEMOCOND_SDL                   (!DEMOCOND_ORIG)

#define GetTicks() ((SDL_GetTicks()*7)/100)

#define ISPOINTER(x) ((((uintptr_t)(x)) & ~0xffff) != 0)

#define CHECKMALLOCRESULT(x) if(!(x)) Quit("Out of memory at %s:%i", __FILE__, __LINE__)

#ifdef _WIN32
    #define strcasecmp stricmp
    #define strncasecmp strnicmp
#else
    static inline char* itoa(int value, char* string, int radix)
    {
	    sprintf(string, "%d", value);
	    return string;
    }

    static inline char* ltoa(long value, char* string, int radix)
    {
	    sprintf(string, "%ld", value);
	    return string;
    }
#endif

#define lengthof(x) (sizeof(x) / sizeof(*(x)))
#define endof(x)    ((x) + lengthof(x))

static inline word READWORD(byte *&ptr)
{
    word val = ptr[0] | ptr[1] << 8;
    ptr += 2;
    return val;
}

static inline longword READLONGWORD(byte *&ptr)
{
    longword val = ptr[0] | ptr[1] << 8 | ptr[2] << 16 | ptr[3] << 24;
    ptr += 4;
    return val;
}

// WSJ: from Matthew's explosion code
static inline fixed GlobalDist(fixed x1, fixed y1, fixed x2, fixed y2)
{
    return (fixed)sqrt(pow(x2-x1,2)+pow(y2-y1,2));
}

/*
=============================================================================

                           FEATURE DEFINITIONS

=============================================================================
*/

#ifdef USE_FEATUREFLAGS
    // The currently available feature flags
    #define FF_STARSKY      0x0001
    #define FF_PARALLAXSKY  0x0002
    #define FF_CLOUDSKY     0x0004
    #define FF_RAIN         0x0010
    #define FF_SNOW         0x0020
    #define FF_POISON       0x0040

    // The ffData... variables contain the 16-bit values of the according corners of the current level.
    // The corners are overwritten with adjacent tiles after initialization in SetupGameLevel
    // to avoid interpretation as e.g. doors.
    extern int ffDataTopLeft, ffDataTopRight, ffDataBottomLeft, ffDataBottomRight;

    /*************************************************************
     * Current usage of ffData... variables:
     * ffDataTopLeft:     lower 8-bit: shade amount; higher 8-bit: shade color
     * ffDataTopRight:    FeatureFlags
     * ffDataBottomLeft:  lower 8-bit: floor color; higher 8-bit: ceiling color or sky
     * ffDataBottomRight: music track (WSJ)
     *************************************************************/

    // The feature flags are stored as a wall in the upper right corner of each level
    static inline word GetFeatureFlags()
    {
        return ffDataTopRight;
    }

#endif

#ifdef USE_FLOORCEILINGTEX
    void DrawFloorAndCeiling(byte *vbuf, unsigned vbufPitch, int min_wallheight);
#endif

#ifdef USE_PARALLAX
    void DrawParallax(byte *vbuf, unsigned vbufPitch);
#endif

#ifdef USE_DIR3DSPR
    void Scale3DShape(byte *vbuf, unsigned vbufPitch, statobj_t *ob);
#endif

#endif
