// WL_STATE.C

#include "wl_def.h"
#pragma hdrstop

/*
=============================================================================

                            LOCAL CONSTANTS

=============================================================================
*/


/*
=============================================================================

                            GLOBAL VARIABLES

=============================================================================
*/


static const dirtype opposite[9] =
    {west,southwest,south,southeast,east,northeast,north,northwest,nodir};

static const dirtype diagonal[9][9] =
{
    /* east */  {nodir,nodir,northeast,nodir,nodir,nodir,southeast,nodir,nodir},
                {nodir,nodir,nodir,nodir,nodir,nodir,nodir,nodir,nodir},
    /* north */ {northeast,nodir,nodir,nodir,northwest,nodir,nodir,nodir,nodir},
                {nodir,nodir,nodir,nodir,nodir,nodir,nodir,nodir,nodir},
    /* west */  {nodir,nodir,northwest,nodir,nodir,nodir,southwest,nodir,nodir},
                {nodir,nodir,nodir,nodir,nodir,nodir,nodir,nodir,nodir},
    /* south */ {southeast,nodir,nodir,nodir,southwest,nodir,nodir,nodir,nodir},
                {nodir,nodir,nodir,nodir,nodir,nodir,nodir,nodir,nodir},
                {nodir,nodir,nodir,nodir,nodir,nodir,nodir,nodir,nodir}
};



void    SpawnNewObj (unsigned tilex, unsigned tiley, statetype *state);
void    NewState (objtype *ob, statetype *state);

boolean TryWalk (objtype *ob);
void    MoveObj (objtype *ob, int32_t move);

void    KillActor (objtype *ob, objtype *killer, damage_t damagetype);
boolean DamageActor (objtype *ob, objtype *attacker, unsigned damage, damage_t damagetype);

boolean CheckLine (objtype *ob);
void    FirstSighting (objtype *ob);
boolean CheckSight (objtype *ob);

/*
=============================================================================

                                LOCAL VARIABLES

=============================================================================
*/



//===========================================================================


/*
===================
=
= SpawnNewObj
=
= Spaws a new actor at the given TILE coordinates, with the given state, and
= the given size in GLOBAL units.
=
= newobj = a pointer to an initialized new actor
=
===================
*/

void SpawnNewObj (unsigned tilex, unsigned tiley, statetype *state)
{
    GetNewActor ();
    newobj->state = state;
    if (state->tictime)
        newobj->ticcount = DEMOCHOOSE_ORIG_SDL(
                US_RndT () % state->tictime,
                US_RndT () % state->tictime + 1);     // Chris' moonwalk bugfix ;D
    else
        newobj->ticcount = 0;

    newobj->tilex = (short) tilex;
    newobj->tiley = (short) tiley;
    newobj->x = ((int32_t)tilex<<TILESHIFT)+TILEGLOBAL/2;
    newobj->y = ((int32_t)tiley<<TILESHIFT)+TILEGLOBAL/2;
    newobj->dir = nodir;

    actorat[tilex][tiley] = newobj;
    newobj->areanumber =
        *(mapsegs[0] + (newobj->tiley<<mapshift)+newobj->tilex) - AREATILE;
}



/*
===================
=
= NewState
=
= Changes ob to a new state, setting ticcount to the max for that state
=
===================
*/

void NewState (objtype *ob, statetype *state)
{
    ob->state = state;
    ob->ticcount = state->tictime;
}



/*
=============================================================================

                        ENEMY TILE WORLD MOVEMENT CODE

=============================================================================
*/


/*
==================================
=
= TryWalk
=
= Attempts to move ob in its current (ob->dir) direction.
=
= If blocked by either a wall or an actor returns FALSE
=
= If move is either clear or blocked only by a door, returns TRUE and sets
=
= ob->tilex         = new destination
= ob->tiley
= ob->areanumber    = the floor tile number (0-(NUMAREAS-1)) of destination
= ob->distance      = TILEGLOBAl, or -doornumber if a door is blocking the way
=
= If a door is in the way, an OpenDoor call is made to start it opening.
= The actor code should wait until
=       doorobjlist[-ob->distance].action = dr_open, meaning the door has been
=       fully opened
=
==================================
*/

boolean CheckEnemyBlock (int x, int y, boolean canshoot)
{
    word tile = MAPSPOT(x,y,1);

    switch (tile)
    {
        case ENEMYBLOCK:
            return canshoot;
        case ENEMYBLOCK2:
        case ACIDTILE:
            return true;
        case KEY1ENEMYBLOCK:
        case KEY2ENEMYBLOCK:
        case KEY3ENEMYBLOCK:
        case KEY4ENEMYBLOCK:
            if ( !(gamestate.keys & (1<<(tile-KEY1ENEMYBLOCK))) )
                return true;
            break;
    }

    return false;
}

#define CHECKDIAG(x,y)                              \
{                                                   \
    if (CheckEnemyBlock(x,y,canshoot))              \
       return false;                                \
    temp=(uintptr_t)actorat[x][y];                  \
    if (temp)                                       \
    {                                               \
        if (temp<256)                               \
            return false;                           \
        if (((objtype *)temp)->flags&FL_SHOOTABLE)  \
            return false;                           \
    }                                               \
}

#ifdef PLAYDEMOLIKEORIGINAL
    #define DOORCHECK                                   \
            if(DEMOCOND_ORIG)                           \
                doornum = temp&63;                      \
            else                                        \
            {                                           \
                doornum = (int) temp & 127;             \
                if (!CheckLock(ob,doornum))             \
                    return false;                       \
                OpenDoor(doornum);                      \
                ob->distance = -doornum - 1;            \
                return true;                            \
            }
#else
    #define DOORCHECK                                   \
            doornum = (int) temp & 127;                 \
            if (!CheckLock(ob,doornum))                 \
                return false;                           \
            OpenDoor(doornum);                          \
            ob->distance = -doornum - 1;                \
            return true;
#endif

#define CHECKSIDE(x,y)                                  \
{                                                       \
    if (CheckEnemyBlock(x,y,canshoot))                  \
       return false;                                    \
    temp=(uintptr_t)actorat[x][y];                      \
    if (temp)                                           \
    {                                                   \
        if (temp<128)                                   \
            return false;                               \
        if (temp<256)                                   \
        {                                               \
            DOORCHECK                                   \
        }                                               \
        else if (((objtype *)temp)->flags&FL_SHOOTABLE) \
            return false;                               \
    }                                                   \
}


boolean TryWalk (objtype *ob)
{
    int       doornum = -1;
    uintptr_t temp;
    boolean   canopen,canshoot;

    switch (ob->obclass)
    {
        case dogobj:
        case ghostobj:
            canopen = canshoot = false;
            break;
        default:
            canopen = canshoot = true;
    }

    if (ob->obclass == inertobj)
    {
        switch (ob->dir)
        {
            case north:
                ob->tiley--;
                break;

            case northeast:
                ob->tilex++;
                ob->tiley--;
                break;

            case east:
                ob->tilex++;
                break;

            case southeast:
                ob->tilex++;
                ob->tiley++;
                break;

            case south:
                ob->tiley++;
                break;

            case southwest:
                ob->tilex--;
                ob->tiley++;
                break;

            case west:
                ob->tilex--;
                break;

            case northwest:
                ob->tilex--;
                ob->tiley--;
                break;
        }
    }
    else
    {
        switch (ob->dir)
        {
            case north:
                if (!canopen)
                {
                    CHECKDIAG(ob->tilex,ob->tiley-1);
                }
                else
                {
                    CHECKSIDE(ob->tilex,ob->tiley-1);
                }
                ob->tiley--;
                break;

            case northeast:
                CHECKDIAG(ob->tilex+1,ob->tiley-1);
                CHECKDIAG(ob->tilex+1,ob->tiley);
                CHECKDIAG(ob->tilex,ob->tiley-1);
                ob->tilex++;
                ob->tiley--;
                break;

            case east:
                if (!canopen)
                {
                    CHECKDIAG(ob->tilex+1,ob->tiley);
                }
                else
                {
                    CHECKSIDE(ob->tilex+1,ob->tiley);
                }
                ob->tilex++;
                break;

            case southeast:
                CHECKDIAG(ob->tilex+1,ob->tiley+1);
                CHECKDIAG(ob->tilex+1,ob->tiley);
                CHECKDIAG(ob->tilex,ob->tiley+1);
                ob->tilex++;
                ob->tiley++;
                break;

            case south:
                if (!canopen)
                {
                    CHECKDIAG(ob->tilex,ob->tiley+1);
                }
                else
                {
                    CHECKSIDE(ob->tilex,ob->tiley+1);
                }
                ob->tiley++;
                break;

            case southwest:
                CHECKDIAG(ob->tilex-1,ob->tiley+1);
                CHECKDIAG(ob->tilex-1,ob->tiley);
                CHECKDIAG(ob->tilex,ob->tiley+1);
                ob->tilex--;
                ob->tiley++;
                break;

            case west:
                if (!canopen)
                {
                    CHECKDIAG(ob->tilex-1,ob->tiley);
                }
                else
                {
                    CHECKSIDE(ob->tilex-1,ob->tiley);
                }
                ob->tilex--;
                break;

            case northwest:
                CHECKDIAG(ob->tilex-1,ob->tiley-1);
                CHECKDIAG(ob->tilex-1,ob->tiley);
                CHECKDIAG(ob->tilex,ob->tiley-1);
                ob->tilex--;
                ob->tiley--;
                break;

            case nodir:
                return false;

            default:
                Quit ("Walk: Bad dir");
        }
    }

#ifdef PLAYDEMOLIKEORIGINAL
    if (DEMOCOND_ORIG && doornum != -1)
    {
        OpenDoor(doornum);
        ob->distance = -doornum-1;
        return true;
    }
#endif

    ob->areanumber =
        *(mapsegs[0] + (ob->tiley<<mapshift)+ob->tilex) - AREATILE;

    ob->distance = TILEGLOBAL;
    return true;
}


/*
==================================
=
= SelectNextDir (by Matthew)
=
= Uses a parallel search algorithm to find and select the next
= direction in as close a path as is possible to the given tile
=
= WSJ: changed to boolean
=
==================================
*/

#define CHECKTILE1(x, y, t) ((!CheckEnemyBlock(x,y,true)) && ((t == 0) \
    || ((t >= 128) && (t < 256)) || (ISPOINTER(actorat[x][y]) && (!(actorat[x][y]->flags & FL_SHOOTABLE)))))

#define CHECKTILE2(x, y, t) ((!CheckEnemyBlock(x,y,false)) && ((t == 0) \
    || (ISPOINTER(actorat[x][y]) && (!(actorat[x][y]->flags & FL_SHOOTABLE)))))

boolean SelectNextDir(objtype *ob, int dtilex, int dtiley)
{
   byte *p1, *p2, *p3;
   byte tx, ty;
   byte tdir;
   int l, t, r, b;
   uintptr_t tmp;

   l = ob->tilex - 1;
   t = ob->tiley - 1;
   r = ob->tilex + 1;
   b = ob->tiley + 1;
   if (ob->tilex == dtilex)
   {
      if (ob->tiley == dtiley)
      {
         ob->dir = nodir;
         return false;
      }
      else if (t == dtiley)
      {
         ob->dir = north;
         if (!TryWalk(ob))
         {
            ob->dir = nodir;
            return false;
         }
         return true;
      }
      else if (b == dtiley)
      {
         ob->dir = south;
         if (!TryWalk(ob))
         {
            ob->dir = nodir;
            return false;
         }
         return true;
      }
   }
   else if (ob->tiley == dtiley)
      if (l == dtilex)
      {
         ob->dir = west;
         if (!TryWalk(ob))
         {
            ob->dir = nodir;
            return false;
         }
         return true;
      }
      else if (r == dtilex)
      {
         ob->dir = east;
         if (!TryWalk(ob))
         {
            ob->dir = nodir;
            return false;
         }
         return true;
      }
      p1 = searchdata;
      p3 = searchdata + ((ob->tiley << mapshift) + ob->tilex);
      for (; p1 < p3; p1++)
         *p1 = 1;
      *(p1++) = 0;
      for (; p1 < sdqstart; p1++)
         *p1 = 1;
      p2 = p1;

      if (ob->obclass == dogobj || ob->obclass == ghostobj)
      {
         if (ob->tiley > 0)
         {
            tx = (byte)ob->tilex;
            ty = (byte)t;
            *(p3 - mapwidth) = 0;
            tmp = (uintptr_t)actorat[tx][ty];
            if (CHECKTILE2(tx, ty, tmp))
            {
               *(p2++) = tx;
               *(p2++) = ty;
               *(p2++) = 2;
            }
         }
         if (ob->tilex < 63)
         {
            tx = (byte)r;
            ty = (byte)ob->tiley;
            *(p3 + 1) = 0;
            tmp = (uintptr_t)actorat[tx][ty];
            if (CHECKTILE2(tx, ty, tmp))
            {
               *(p2++) = tx;
               *(p2++) = ty;
               *(p2++) = 0;
            }
         }
         if (ob->tiley < 63)
         {
            tx = (byte)ob->tilex;
            ty = (byte)b;
            *(p3 + mapwidth) = 0;
            tmp = (uintptr_t)actorat[tx][ty];
            if (CHECKTILE2(tx, ty, tmp))
            {
               *(p2++) = tx;
               *(p2++) = ty;
               *(p2++) = 6;
            }
         }
         if (ob->tilex > 0)
         {
            tx = (byte)l;
            ty = (byte)ob->tiley;
            *(p3 - 1) = 0;
            tmp = (uintptr_t)actorat[tx][ty];
            if (CHECKTILE2(tx, ty, tmp))
            {
               *(p2++) = tx;
               *(p2++) = ty;
               *(p2++) = 4;
            }
         }
         for (; p1 < p2; )
         {
            tx = *(p1++);
            ty = *(p1++);
            l = tx - 1;
            t = ty - 1;
            r = tx + 1;
            b = ty + 1;
            if (tx == dtilex)
            {
               if ((t == dtiley) || (b == dtiley))
               {
                  ob->dir = (dirtype)*p1;
                  if (!TryWalk(ob))
                  {
                     ob->dir = nodir;
                     return false;
                  }
                  return true;
               }
            } else if (ty == dtiley)
               if ((l == dtilex) || (r == dtilex))
               {
                  ob->dir = (dirtype)*p1;
                  if (!TryWalk(ob))
                  {
                     ob->dir = nodir;
                     return false;
                  }
                  return true;
               }
            tdir = *(p1++);
            if (ty > 0)
            {
               p3 = searchdata + ((t << mapshift) + (int)tx);
               if (*p3)
               {
                  *p3 = 0;
                  tmp = (uintptr_t)actorat[tx][t];
                  if (CHECKTILE2(tx, t, tmp))
                  {
                     *(p2++) = tx;
                     *(p2++) = (byte)t;
                     *(p2++) = tdir;
                  }
               }
            }
            if (tx < 63)
            {
               p3 = searchdata + (((int)ty << mapshift) + r);
               if (*p3)
               {
                  *p3 = 0;
                  tmp = (uintptr_t)actorat[r][ty];
                  if (CHECKTILE2(r, ty, tmp))
                  {
                     *(p2++) = (byte)r;
                     *(p2++) = ty;
                     *(p2++) = tdir;
                  }
               }
            }
            if (ty < 63)
            {
               p3 = searchdata + ((b << mapshift) + (int)tx);
               if (*p3)
               {
                  tmp = (uintptr_t)actorat[tx][b];
                  if (CHECKTILE2(tx, b, tmp))
                  {
                     *(p2++) = tx;
                     *(p2++) = (byte)b;
                     *(p2++) = tdir;
                  }
               }
            }
            if (tx > 0)
            {
               p3 = searchdata + (((int)ty << mapshift) + l);
               if (*p3)
               {
                  *p3 = 0;
                  tmp = (uintptr_t)actorat[l][ty];
                  if (CHECKTILE2(l, ty, tmp))
                  {
                     *(p2++) = (byte)l;
                     *(p2++) = ty;
                     *(p2++) = tdir;
                  }
               }
            }
         }
      }
      else
      {
         if (ob->tiley > 0)
         {
            tx = (byte)ob->tilex;
            ty = (byte)t;
            *(p3 - mapwidth) = 0;
            tmp = (uintptr_t)actorat[tx][ty];
            if (CHECKTILE1(tx, ty, tmp))
            {
               *(p2++) = tx;
               *(p2++) = ty;
               *(p2++) = 2;
            }
         }
         if (ob->tilex < 63)
         {
            tx = (byte)r;
            ty = (byte)ob->tiley;
            *(p3 + 1) = 0;
            tmp = (uintptr_t)actorat[tx][ty];
            if (CHECKTILE1(tx, ty, tmp))
            {
               *(p2++) = tx;
               *(p2++) = ty;
               *(p2++) = 0;
            }
         }
         if (ob->tiley < 63)
         {
            tx = (byte)ob->tilex;
            ty = (byte)b;
            *(p3 + mapwidth) = 0;
            tmp = (uintptr_t)actorat[tx][ty];
            if (CHECKTILE1(tx, ty, tmp))
            {
               *(p2++) = tx;
               *(p2++) = ty;
               *(p2++) = 6;
            }
         }
         if (ob->tilex > 0)
         {
            tx = (byte)l;
            ty = (byte)ob->tiley;
            *(p3 - 1) = 0;
            tmp = (uintptr_t)actorat[tx][ty];
            if (CHECKTILE1(tx, ty, tmp))
            {
               *(p2++) = tx;
               *(p2++) = ty;
               *(p2++) = 4;
            }
         }
         for (; p1 < p2; )
         {
            tx = *(p1++);
            ty = *(p1++);
            l = tx - 1;
            t = ty - 1;
            r = tx + 1;
            b = ty + 1;
            if (tx == dtilex)
            {
               if ((t == dtiley) || (b == dtiley))
               {
                  ob->dir = (dirtype)*p1;
                  if (!TryWalk(ob))
                  {
                     ob->dir = nodir;
                     return false;
                  }
                  return true;
               }
            } else if (ty == dtiley)
               if ((l == dtilex) || (r == dtilex))
               {
                  ob->dir = (dirtype)*p1;
                  if (!TryWalk(ob))
                  {
                     ob->dir = nodir;
                     return false;
                  }
                  return true;
               }
            tdir = *(p1++);
            if (ty > 0)
            {
               p3 = searchdata + ((t << mapshift) + (int)tx);
               if (*p3)
               {
                  *p3 = 0;
                  tmp = (uintptr_t)actorat[tx][t];
                  if (CHECKTILE1(tx, t, tmp))
                  {
                     *(p2++) = tx;
                     *(p2++) = (byte)t;
                     *(p2++) = tdir;
                  }
               }
            }
            if (tx < 63)
            {
               p3 = searchdata + (((int)ty << mapshift) + r);
               if (*p3)
               {
                  *p3 = 0;
                  tmp = (uintptr_t)actorat[r][ty];
                  if (CHECKTILE1(r, ty, tmp))
                  {
                     *(p2++) = (byte)r;
                     *(p2++) = ty;
                     *(p2++) = tdir;
                  }
               }
            }
            if (ty < 63)
            {
               p3 = searchdata + ((b << mapshift) + (int)tx);
               if (*p3)
               {
                  tmp = (uintptr_t)actorat[tx][b];
                  if (CHECKTILE1(tx, b, tmp))
                  {
                     *(p2++) = tx;
                     *(p2++) = (byte)b;
                     *(p2++) = tdir;
                  }
               }
            }
            if (tx > 0)
            {
               p3 = searchdata + (((int)ty << mapshift) + l);
               if (*p3)
               {
                  *p3 = 0;
                  tmp = (uintptr_t)actorat[l][ty];
                  if (CHECKTILE1(l, ty, tmp))
                  {
                     *(p2++) = (byte)l;
                     *(p2++) = ty;
                     *(p2++) = tdir;
                  }
               }
            }
         }
      }
      ob->dir = nodir;
      return false;
}


/*
==================================
=
= SelectDodgeDir
=
= Attempts to choose and initiate a movement for ob that sends it towards
= the player while dodging
=
= If there is no possible move (ob is totally surrounded)
=
= ob->dir           =       nodir
=
= Otherwise
=
= ob->dir           = new direction to follow
= ob->distance      = TILEGLOBAL or -doornumber
= ob->tilex         = new destination
= ob->tiley
= ob->areanumber    = the floor tile number (0-(NUMAREAS-1)) of destination
=
==================================
*/

void SelectDodgeDir (objtype *ob)
{
    int         deltax,deltay,i;
    unsigned    absdx,absdy;
    dirtype     dirtry[5];
    dirtype     turnaround,tdir;


    if (ob->flags & FL_FIRSTATTACK)
    {
        //
        // turning around is only ok the very first time after noticing the
        // player
        //
        turnaround = nodir;
        ob->flags &= ~FL_FIRSTATTACK;
    }
    else
        turnaround=opposite[ob->dir];

    deltax = player->tilex - ob->tilex;
    deltay = player->tiley - ob->tiley;

    //
    // arange 5 direction choices in order of preference
    // the four cardinal directions plus the diagonal straight towards
    // the player
    //

    if (deltax>0)
    {
        dirtry[1]= east;
        dirtry[3]= west;
    }
    else
    {
        dirtry[1]= west;
        dirtry[3]= east;
    }

    if (deltay>0)
    {
        dirtry[2]= south;
        dirtry[4]= north;
    }
    else
    {
        dirtry[2]= north;
        dirtry[4]= south;
    }

    //
    // randomize a bit for dodging
    //
    absdx = abs(deltax);
    absdy = abs(deltay);

    if (absdx > absdy)
    {
        tdir = dirtry[1];
        dirtry[1] = dirtry[2];
        dirtry[2] = tdir;
        tdir = dirtry[3];
        dirtry[3] = dirtry[4];
        dirtry[4] = tdir;
    }

    if (US_RndT() < 128)
    {
        tdir = dirtry[1];
        dirtry[1] = dirtry[2];
        dirtry[2] = tdir;
        tdir = dirtry[3];
        dirtry[3] = dirtry[4];
        dirtry[4] = tdir;
    }

    dirtry[0] = diagonal [ dirtry[1] ] [ dirtry[2] ];

    //
    // try the directions util one works
    //
    for (i=0;i<5;i++)
    {
        if ( dirtry[i] == nodir || dirtry[i] == turnaround)
            continue;

        ob->dir = dirtry[i];
        if (TryWalk(ob))
            return;
    }

    //
    // turn around only as a last resort
    //
    if (turnaround != nodir)
    {
        ob->dir = turnaround;

        if (TryWalk(ob))
            return;
    }

    ob->dir = nodir;
}


/*
============================
=
= SelectChaseDir
=
= As SelectDodgeDir, but doesn't try to dodge
=
============================
*/

void SelectChaseDir (objtype *ob)
{
    int     deltax,deltay;
    dirtype d[3];
    dirtype tdir, olddir, turnaround;

    if (SelectNextDir(ob, (byte)player->tilex, (byte)player->tiley))
        return;    // WSJ: use advanced search instead

    olddir=ob->dir;
    turnaround=opposite[olddir];

    deltax=player->tilex - ob->tilex;
    deltay=player->tiley - ob->tiley;

    d[1]=nodir;
    d[2]=nodir;

    if (deltax>0)
        d[1]= east;
    else if (deltax<0)
        d[1]= west;
    if (deltay>0)
        d[2]=south;
    else if (deltay<0)
        d[2]=north;

    if (abs(deltay)>abs(deltax))
    {
        tdir=d[1];
        d[1]=d[2];
        d[2]=tdir;
    }

    if (d[1]==turnaround)
        d[1]=nodir;
    if (d[2]==turnaround)
        d[2]=nodir;


    if (d[1]!=nodir)
    {
        ob->dir=d[1];
        if (TryWalk(ob))
            return;     /*either moved forward or attacked*/
    }

    if (d[2]!=nodir)
    {
        ob->dir=d[2];
        if (TryWalk(ob))
            return;
    }

    /* there is no direct path to the player, so pick another direction */

    if (olddir!=nodir)
    {
        ob->dir=olddir;
        if (TryWalk(ob))
            return;
    }

    if (US_RndT()>128)      /*randomly determine direction of search*/
    {
        for (tdir=north; tdir<=west; tdir=(dirtype)(tdir+1))
        {
            if (tdir!=turnaround)
            {
                ob->dir=tdir;
                if ( TryWalk(ob) )
                    return;
            }
        }
    }
    else
    {
        for (tdir=west; tdir>=north; tdir=(dirtype)(tdir-1))
        {
            if (tdir!=turnaround)
            {
                ob->dir=tdir;
                if ( TryWalk(ob) )
                    return;
            }
        }
    }

    if (turnaround !=  nodir)
    {
        ob->dir=turnaround;
        if (ob->dir != nodir)
        {
            if ( TryWalk(ob) )
                return;
        }
    }

    ob->dir = nodir;                // can't move
}

/*
============================
=
= SelectRunDir
=
= Run Away from player
=
============================
*/

void SelectRunDir (objtype *ob)
{
    int deltax,deltay;
    dirtype d[3];
    dirtype tdir;


    deltax=player->tilex - ob->tilex;
    deltay=player->tiley - ob->tiley;

    if (deltax<0)
        d[1]= east;
    else
        d[1]= west;
    if (deltay<0)
        d[2]=south;
    else
        d[2]=north;

    if (abs(deltay)>abs(deltax))
    {
        tdir=d[1];
        d[1]=d[2];
        d[2]=tdir;
    }

    ob->dir=d[1];
    if (TryWalk(ob))
        return;     /*either moved forward or attacked*/

    ob->dir=d[2];
    if (TryWalk(ob))
        return;

    /* there is no direct path to the player, so pick another direction */

    if (US_RndT()>128)      /*randomly determine direction of search*/
    {
        for (tdir=north; tdir<=west; tdir=(dirtype)(tdir+1))
        {
            ob->dir=tdir;
            if ( TryWalk(ob) )
                return;
        }
    }
    else
    {
        for (tdir=west; tdir>=north; tdir=(dirtype)(tdir-1))
        {
            ob->dir=tdir;
            if ( TryWalk(ob) )
                return;
        }
    }

    ob->dir = nodir;                // can't move
}


/*
=================
=
= MoveObj
=
= Moves ob be move global units in ob->dir direction
= Actors are not allowed to move inside the player
= Does NOT check to see if the move is tile map valid
=
= ob->x                 = adjusted for new position
= ob->y
=
=================
*/

void MoveObj (objtype *ob, int32_t move)
{
    MoveObjInDir (ob,move,ob->dir);
}

void MoveObjInDir (objtype *ob, int32_t move, dirtype dir)
{
    int32_t    deltax,deltay;

    switch (dir)
    {
        case north:
            ob->y -= move;
            break;
        case northeast:
            ob->x += move;
            ob->y -= move;
            break;
        case east:
            ob->x += move;
            break;
        case southeast:
            ob->x += move;
            ob->y += move;
            break;
        case south:
            ob->y += move;
            break;
        case southwest:
            ob->x -= move;
            ob->y += move;
            break;
        case west:
            ob->x -= move;
            break;
        case northwest:
            ob->x -= move;
            ob->y -= move;
            break;

        case nodir:
            return;

        default:
            Quit ("MoveObj: bad dir!");
    }

    //
    // check to make sure it's not on top of player
    //
    if (areabyplayer[ob->areanumber])
    {
        deltax = ob->x - player->x;
        if (deltax < -MINACTORDIST || deltax > MINACTORDIST)
            goto moveok;
        deltay = ob->y - player->y;
        if (deltay < -MINACTORDIST || deltay > MINACTORDIST)
            goto moveok;

        if (ob->hidden)          // move closer until he meets CheckLine
            goto moveok;

        if (ob->obclass == ghostobj)
            TakeDamage (tics*2,ob,dam_melee,true,true);

        //
        // back up
        //
        switch (dir)
        {
            case north:
                ob->y += move;
                break;
            case northeast:
                ob->x -= move;
                ob->y += move;
                break;
            case east:
                ob->x -= move;
                break;
            case southeast:
                ob->x -= move;
                ob->y -= move;
                break;
            case south:
                ob->y -= move;
                break;
            case southwest:
                ob->x += move;
                ob->y -= move;
                break;
            case west:
                ob->x += move;
                break;
            case northwest:
                ob->x += move;
                ob->y += move;
                break;

            case nodir:
                return;
        }
        return;
    }
moveok:
    ob->distance -=move;
}

/*
=============================================================================

                                STUFF

=============================================================================
*/

/*
===============
=
= DropItem
=
===============
*/

void DropItem (objtype *ob)
{
    word tilex = ob->tilex = (word)(ob->x >> TILESHIFT);
    word tiley = ob->tiley = (word)(ob->y >> TILESHIFT);

    switch (ob->obclass)
    {
        case guardobj:
        case officerobj:
        case ssofficer2obj:
        case scientistobj:
            PlaceItemType(bo_pistol,tilex,tiley);
            break;
        case infantryobj:
        case officer2obj:
        case ssobj:
        case ssofficerobj:
        case mutantobj:
            PlaceItemType(bo_mp40,tilex,tiley);
            break;
        case riflemanobj:
        case ssriflemanobj:
            PlaceItemType(bo_kar98k,tilex,tiley);
            break;
        case sniperobj:
            PlaceItemType(bo_kar98k_sniper,tilex,tiley);
            break;
        case waffenssobj:
        case waffenofficerobj:
            PlaceItemType(bo_stg44,tilex,tiley);
            break;
        case mg42guardobj:
        case mg42ssobj:
        case mg42waffenssobj:
            PlaceItemType(bo_mg42_mounted,tilex,tiley);
            break;
        case mechtrooperobj:
            PlaceItemType(bo_mg42,tilex,tiley);
            break;
        case flametrooperobj:
            PlaceItemType(bo_flamethrower,tilex,tiley);
            break;
        case chaingunnerobj:
            PlaceItemType(bo_chaingun,tilex,tiley);
            break;
        case rockettrooperobj:
            PlaceItemType(bo_panzerschreck,tilex,tiley);
            break;
        case bossobj:
        case schabbobj:
            PlaceItemType(bo_key1,tilex,tiley);
            break;
        case fatobj:
            PlaceItemType(bo_key3,tilex,tiley);
            break;
        case panzobj:
            PlaceItemType(bo_key2,tilex,tiley);
            break;
    }
}


/*
===============
=
= KillActor
=
===============
*/

void KillActor (objtype *ob, objtype *killer, damage_t damagetype)
{
    int killpoints,deathtype;

    if (damagetype == dam_rocket && ob->hitpoints < -25 && OBJTILEDIST(ob,killer) < 3)
        deathtype = 1;    // WSJ: gib death
    else if (damagetype >= dam_chaingun && Random(4)==3 && OBJTILEDIST(ob,killer) < 3)
        deathtype = 1;    // WSJ: gib death
    else if (damagetype == dam_fire)
        deathtype = 2;    // WSJ: burning death
    else
        deathtype = 0;    // WSJ: default death

    actorat[ob->tilex][ob->tiley] = NULL;
    ob->active = ac_yes;
    ob->flags |= FL_NONMARK;

    killpoints = 0;

    switch (ob->obclass)
    {
        case officerobj:
        case officer2obj:
        case mg42guardobj:
            killpoints += 300;
        case guardobj:
        case infantryobj:
        case riflemanobj:
            killpoints += 100;
            switch (deathtype)
            {
                case 1:
                    ob->sproffset = 0;
                    NewState (ob,&s_grdgib1);
                    break;
                case 2:
                    ob->sproffset = 0;
                    NewState (ob,&s_grdburn1);
                    break;
                default:
                    NewState (ob,&s_grddie1);
            }
            break;

        case mutantobj:
            killpoints += 700;
            switch (deathtype)
            {
                case 1:
                    ob->sproffset = SPR_MUT_GIB_1-SPR_GRD_GIB_1;
                    NewState (ob,&s_grdgib1);
                    break;
                case 2:
                    ob->sproffset = 0;
                    NewState (ob,&s_mutburn1);
                    break;
                default:
                    NewState (ob,&s_grddie1);
            }
            break;

        case ssofficerobj:
        case ssofficer2obj:
        case mg42ssobj:
        case sniperobj:
            killpoints += 300;
        case ssobj:
        case ssriflemanobj:
            killpoints += 500;
            switch (deathtype)
            {
                case 1:
                    ob->sproffset = SPR_SS_GIB_1-SPR_GRD_GIB_1;
                    NewState (ob,&s_grdgib1);
                    break;
                case 2:
                    ob->sproffset = 0;
                    NewState (ob,&s_grdburn1);
                    break;
                default:
                    NewState (ob,&s_grddie1);
            }
            break;

        case dogobj:
            killpoints += 200;
            NewState (ob,&s_dogdie1);
            break;

        case scientistobj:
            killpoints += 400;
            switch (deathtype)
            {
                case 1:
                    ob->sproffset = SPR_SCI_GIB_1-SPR_GRD_GIB_1;
                    NewState (ob,&s_grdgib1);
                    break;
                case 2:
                    ob->sproffset = 0;
                    NewState (ob,&s_grdburn1);
                    break;
                default:
                    NewState (ob,&s_grddie1);
            }
            break;

        case waffenofficerobj:
        case mg42waffenssobj:
            killpoints += 200;
        case waffenssobj:
            killpoints += 700;
            switch (deathtype)
            {
                case 1:
                    ob->sproffset = SPR_WAFFEN_GIB_1-SPR_GRD_GIB_1;
                    NewState (ob,&s_grdgib1);
                    break;
                case 2:
                    ob->sproffset = 0;
                    NewState (ob,&s_grdburn1);
                    break;
                default:
                    NewState (ob,&s_grddie1);
            }
            break;

        case mechtrooperobj:
            killpoints += 2000;
            NewState (ob,&s_mstdie1);
            break;

        case bossobj:
        case schabbobj:
            killpoints += 3000;
        case flametrooperobj:
        case chaingunnerobj:
        case rockettrooperobj:
            killpoints += 2000;
            NewState (ob,&s_bossdie1);
            break;

        case fatobj:
            killpoints += 5000;
            NewState (ob,&s_narbdie1);
            break;

        case deatharmorobj:
            killpoints += 5000;
            NewState (ob,&s_mechadie1);
            break;

        case panzobj:
            killpoints += 5000;
            NewState (ob,&s_panzdie1);
            break;
    }

    ob->alert = alert_dead;
    ob->flags &= ~FL_SHOOTABLE;
    ob->flags &= ~FL_ATTACKMODE;
    DropItem(ob);

    if ( !(ob->flags & FL_BONUS) )
        return;    // WSJ: don't count resurrected enemy

    ob->flags &= ~FL_BONUS;
    gamestate.killcount++;
    GivePoints(killpoints);
}



/*
===================
=
= DamageActor
=
= Called when the player succesfully hits an enemy.
=
= Does damage points to enemy ob, either putting it into a stun frame or
= killing it.
=
= WSJ: changed to a boolean in case the enemy was immune to the damage
=
===================
*/

boolean DamageActor (objtype *ob, objtype *attacker, unsigned damage, damage_t damagetype)
{
    madenoise = true;

    if (ob->alert == alert_dormant)
        return false;

    switch (ob->obclass)
    {
        case mechtrooperobj:
            if (damagetype < dam_bullet || damage<20)
                return false;
            break;
        case bossobj:
        case deatharmorobj:
            if (damagetype < dam_bullet || damage<25)
                return false;
            break;
        case panzobj:
            if (damagetype < dam_flak)
                return false;
            break;
        case fatobj:
        case schabbobj:
            break;
        default:
            if (damagetype >= dam_flak)
                damage *= 2;
    }

    //
    // do double damage if shooting a non attack mode actor
    //
    if ( !(ob->flags & FL_ATTACKMODE) )
        damage <<= 1;

    ob->hitpoints -= (short)damage;

    if (ob->hitpoints<=0)
        KillActor (ob,attacker,damagetype);
    else
    {
        if ( !(ob->flags & FL_ATTACKMODE) )
            FirstSighting (ob);             // put into combat mode

        switch (ob->obclass)                // dogs only have one hit point
        {
            case guardobj:
            case riflemanobj:
            case infantryobj:
            case officerobj:
            case officer2obj:
            case ssobj:
            case ssriflemanobj:
            case ssofficerobj:
            case ssofficer2obj:
            case scientistobj:
            case waffenssobj:
            case waffenofficerobj:
            case mutantobj:
                if (ob->hitpoints&1)
                    NewState (ob,&s_grdpain);
                else
                    NewState (ob,&s_grdpain1);
                break;

            case sniperobj:
            case mg42guardobj:
            case mg42ssobj:
            case mg42waffenssobj:
                if (ob->hitpoints&1)
                    NewState (ob,&s_sniperpain);
                else
                    NewState (ob,&s_sniperpain1);
                break;
        }
    }

    return true;
}

/*
=============================================================================

                                CHECKSIGHT

=============================================================================
*/


/*
=====================
=
= CheckLine
=
= Returns true if a straight line between the player and ob is unobstructed
=
= WSJ: edited to check between specific coordinates
=
=====================
*/

boolean CheckCoordLine (fixed fx1, fixed fy1, fixed fx2, fixed fy2)
{
    int         x1,y1,xt1,yt1,x2,y2,xt2,yt2;
    int         x,y;
    int         xdist,ydist,xstep,ystep;
    int         partial,delta;
    int32_t     ltemp;
    int         xfrac,yfrac,deltafrac;
    unsigned    value,intercept;


    x1 = fx1 >> UNSIGNEDSHIFT;            // 1/256 tile precision
    y1 = fy1 >> UNSIGNEDSHIFT;
    xt1 = x1 >> 8;
    yt1 = y1 >> 8;

    x2 = fx2 >> UNSIGNEDSHIFT;
    y2 = fy2 >> UNSIGNEDSHIFT;
    xt2 = x2 >> 8;
    yt2 = y2 >> 8;

    xdist = abs(xt2-xt1);

    if (xdist > 0)
    {
        if (xt2 > xt1)
        {
            partial = 256-(x1&0xff);
            xstep = 1;
        }
        else
        {
            partial = x1&0xff;
            xstep = -1;
        }

        deltafrac = abs(x2-x1);
        delta = y2-y1;
        ltemp = ((int32_t)delta<<8)/deltafrac;
        if (ltemp > 0x7fffl)
            ystep = 0x7fff;
        else if (ltemp < -0x7fffl)
            ystep = -0x7fff;
        else
            ystep = ltemp;
        yfrac = y1 + (((int32_t)ystep*partial) >>8);

        x = xt1+xstep;
        xt2 += xstep;
        do
        {
            y = yfrac>>8;
            yfrac += ystep;

            value = (unsigned)tilemap[x][y];
            x += xstep;

            if (!value)
                continue;

            if (value<128 || value>256)
                return false;

            //
            // see if the door is open enough
            //
            value &= ~0x80;
            intercept = yfrac-ystep/2;

            if (intercept>doorposition[value])
                return false;

        } while (x != xt2);
    }

    ydist = abs(yt2-yt1);

    if (ydist > 0)
    {
        if (yt2 > yt1)
        {
            partial = 256-(y1&0xff);
            ystep = 1;
        }
        else
        {
            partial = y1&0xff;
            ystep = -1;
        }

        deltafrac = abs(y2-y1);
        delta = x2-x1;
        ltemp = ((int32_t)delta<<8)/deltafrac;
        if (ltemp > 0x7fffl)
            xstep = 0x7fff;
        else if (ltemp < -0x7fffl)
            xstep = -0x7fff;
        else
            xstep = ltemp;
        xfrac = x1 + (((int32_t)xstep*partial) >>8);

        y = yt1 + ystep;
        yt2 += ystep;
        do
        {
            x = xfrac>>8;
            xfrac += xstep;

            value = (unsigned)tilemap[x][y];
            y += ystep;

            if (!value)
                continue;

            if (value<128 || value>256)
                return false;

            //
            // see if the door is open enough
            //
            value &= ~0x80;
            intercept = xfrac-xstep/2;

            if (intercept>doorposition[value])
                return false;
        } while (y != yt2);
    }

    return true;
}

boolean CheckLine (objtype *ob)
{
    return CheckCoordLine (ob->x,ob->y,player->x,player->y);
}

boolean CheckTileLine (unsigned tx1, unsigned ty1, unsigned tx2, unsigned ty2)
{
    int32_t x1,y1,x2,y2;

    x1 = ((int32_t)tx1<<TILESHIFT)+TILEGLOBAL/2;
    y1 = ((int32_t)ty1<<TILESHIFT)+TILEGLOBAL/2;
    x2 = ((int32_t)tx2<<TILESHIFT)+TILEGLOBAL/2;
    y2 = ((int32_t)ty2<<TILESHIFT)+TILEGLOBAL/2;

    return CheckCoordLine (x1,y1,x2,y2);
}


/*
================
=
= CheckSight
=
= Checks a straight line between player and current object
=
= If the sight is ok, check alertness and angle to see if they notice
=
= returns true if the player has been spoted
=
================
*/

#define MINSIGHT        0x18000l

boolean CheckObjSight (objtype *ob, objtype *check)
{
    int32_t deltax,deltay;

    //
    // don't bother tracing a line if the area isn't connected to the player's
    //
    if (!areabyplayer[ob->areanumber])
        return false;

    //
    // if the player is real close, sight is automatic
    //
    deltax = check->x - ob->x;
    deltay = check->y - ob->y;

    if (deltax > -MINSIGHT && deltax < MINSIGHT
        && deltay > -MINSIGHT && deltay < MINSIGHT)
        return true;

    //
    // see if they are looking in the right direction
    //
    switch (ob->dir)
    {
        case north:
            if (deltay > 0)
                return false;
            break;

        case east:
            if (deltax < 0)
                return false;
            break;

        case south:
            if (deltay < 0)
                return false;
            break;

        case west:
            if (deltax > 0)
                return false;
            break;

        // check diagonal moving guards fix

        case northwest:
            if (DEMOCOND_SDL && deltay > -deltax)
                return false;
            break;

        case northeast:
            if (DEMOCOND_SDL && deltay > deltax)
                return false;
            break;

        case southwest:
            if (DEMOCOND_SDL && deltax > deltay)
                return false;
            break;

        case southeast:
            if (DEMOCOND_SDL && -deltax > deltay)
                return false;
            break;
    }

    //
    // trace a line to check for blocking tiles (corners)
    //
    return CheckObjLine(ob,check);
}

boolean CheckSight (objtype *ob)
{
    objtype *check;

    if (ob->flags & FL_AMBUSH)
        return CheckObjSight(ob,player);

    // WSJ: check for dead bodies and alerted guards
    for (check = player; check; check = check->next)
    {
        if (check != player && !(check->flags & FL_ATTACKMODE) && check->alert != alert_dead)
            continue;

        if (CheckObjSight(ob,check))
        {
            if (check == player)
                ob->alert = alert_sight;
            else if (check->alert == alert_dead)
                ob->alert = alert_findbody;

            return true;
        }
    }

    return false;
}


/*
===============
=
= FirstSighting
=
= Puts an actor into attack mode and possibly reverses the direction
= if the player is behind it
=
===============
*/

void FirstSighting (objtype *ob)
{
    //
    // react to the player
    //
    switch (ob->obclass)
    {
        case guardobj:
        case riflemanobj:
        case infantryobj:
        case ssobj:
        case ssriflemanobj:
        case waffenssobj:
            NewState (ob,&s_grdchase1);
            ob->speed *= 4;                 // go faster when chasing player
            break;

        case officerobj:
        case officer2obj:
        case ssofficerobj:
        case ssofficer2obj:
        case scientistobj:
        case waffenofficerobj:
            NewState (ob,&s_grdchase1);
            ob->speed *= 5;                 // go faster when chasing player
            break;

        case mutantobj:
            NewState (ob,&s_grdchase1);
            ob->speed *= 3;                 // go faster when chasing player
            break;

        case dogobj:
            NewState (ob,&s_dogchase1);
            ob->speed *= 2;                 // go faster when chasing player
            break;

        case mechtrooperobj:
            NewState (ob,&s_mstchase1);
            ob->speed *= 3;                 // go faster when chasing player
            break;

        case bossobj:
        case deatharmorobj:
        case chaingunnerobj:
        case rockettrooperobj:
            NewState (ob,&s_bosschase1);
            ob->speed = SPDPATROL*3;        // go faster when chasing player
            break;

        case flametrooperobj:
            NewState (ob,&s_bosschase1);
            ob->speed = SPDPATROL*4;        // go faster when chasing player
            break;

        case fatobj:
            NewState (ob,&s_bosschase1);
            ob->speed *= 5;                 // go faster when chasing player
            break;

        case schabbobj:
            //NewState (ob,&s_bosschase1);
            NewState (ob,&s_schabbshoot1);
            ob->speed *= 5;                 // go faster when chasing player
            break;

        case panzobj:
            NewState (ob,&s_bossshoot1);
            ob->speed *= 3;                 // go faster when chasing player
            break;
    }

    if (ob->distance < 0)
        ob->distance = 0;       // ignore the door opening command

    ob->flags |= FL_ATTACKMODE;
    ob->flags |= FL_FIRSTATTACK;
    ob->active = ac_yes;
    if (CheckBoss(ob))
        gamestate.startbosshp += ob->hitpoints;
}


/*
===============
=
= SightPlayer
=
= Called by actors that ARE NOT chasing the player.  If the player
= is detected (by sight, noise, or proximity), the actor is put into
= it's combat frame and true is returned.
=
= Incorporates a random reaction delay
=
===============
*/

boolean SightPlayer (objtype *ob)
{
    if (ob->flags & FL_ATTACKMODE)
        Quit ("An actor in ATTACKMODE called SightPlayer!");

    if (invisiblemode)
        return false;

    if (ob->temp2)
    {
        //
        // count down reaction time
        //
        ob->temp2 -= (short) tics;
        if (ob->temp2 > 0)
            return false;
        ob->temp2 = 0;                                  // time to react
    }
    else
    {
        if (ob->alert == alert_dormant)
        {
            if (!areabyplayer[ob->areanumber])
                return false;

            ob->alert = alert_none;
        }
        else if (!CheckSight (ob))
        {
            if (ob->flags & FL_AMBUSH)
                return false;

            if (!AreaByActor(ob,lastalertedarea))
                return false;

            if (madenoise)
                ob->alert = alert_hear;
        }

        if (ob->flags & FL_AMBUSH)
        {
            ob->temp2 = 1;    // WSJ: instant reaction by default
        }
        else
        {
            switch (ob->obclass)
            {
            case guardobj:
            case riflemanobj:
            case infantryobj:
                ob->temp2 = 1+US_RndT()/4;
                break;
            case officerobj:
            case officer2obj:
            case ssofficerobj:
            case ssofficer2obj:
            case scientistobj:
            case waffenofficerobj:
                ob->temp2 = 2;
                break;
            case mutantobj:
            case ssobj:
            case ssriflemanobj:
            case waffenssobj:
                ob->temp2 = 1+US_RndT()/6;
                break;
            case dogobj:
                ob->temp2 = 1+US_RndT()/8;
                break;
            default:
                ob->temp2 = 1;    // WSJ: instant reaction by default
            }
        }
        return false;
    }

    A_AlertSound (ob);

    FirstSighting (ob);

    return true;
}
